import os
from dataclasses import dataclass
from typing import List

import pytest

from tecton.cli.environment import EnvironmentDependencies
from tecton.cli.environment import UploadPart
from tecton.cli.environment import _get_pkg_to_version
from tecton.cli.environment import _run_dependency_resolution
from tecton.cli.environment import get_upload_parts
from tecton.cli.environment_utils import is_valid_environment_name
from tecton.cli.upload_utils import DEFAULT_UPLOAD_PART_SIZE_MB


ERROR_INPUT_REQUIREMENTS_TEXT = """
urllib3<1.27
tecton==1.1.0b6
"""

INPUT_REQUIREMENTS_TEXT = """
# THIS IS A MESS
# This is an implicit value, here for clarity
--index-url https://pypi.python.org/simple/

# pypika @ https://s3.us-west-2.amazonaws.com/tecton.ai.public/python/PyPika-0.48.9-py2.py3-none-any.whl
xformers==0.0.28.post1
# tecton[rift-materialization] @ file:///Users/vitaly/dev/tecton/bazel-bin/sdk/pypi/tecton-99.99.99-py3-none-any.whl # hello
tecton-runtime==1.0.0

# pydantic<2
# setuptools<70
urllib3<1.27

# protobuf<=4.25.3 # hi
tecton[rift-materialization]==1.1.0b6
"""

RESOLVED_REQUIREMENTS_TEXT = """
--index-url https://pypi.python.org/simple/
--find-links https://s3.us-west-2.amazonaws.com/tecton.ai.public/python/index.html

deltalake==0.18.2
    # via tecton
    # from https://pypi.python.org/simple/
tecton[rift-materialization] @ file:///Users/vitaly/dev/tecton/bazel-bin/sdk/pypi/tecton-99.99.99-py3-none-any.whl
    # via -r requirements.txt
tecton-runtime==1.0.0
    # via -r requirements.txt
    # from https://pypi.python.org/simple/
"""


SAMPLE_LOCK_JSON = {
    "locked_resolves": [
        {
            "locked_requirements": [
                {
                    "project_name": "attrs",
                    "requires_dists": [
                        "attrs[tests-mypy]; extra == 'tests-no-zope'",
                        "hypothesis; extra == 'tests-no-zope'",
                    ],
                    "version": "23.2.0",
                },
                {"project_name": "boto3", "requires_dists": ["botocore<1.35.0,>=1.34.87"], "version": "1.34.87"},
                {"project_name": "pytest", "requires_dists": [], "version": "8.1.1"},
            ]
        }
    ],
    "requirements": ["attrs==23.2.0", "boto3==1.34.87", "pytest"],
}


@pytest.fixture
def uv_binary_setup(tmp_path):
    import importlib.metadata
    from importlib.util import find_spec

    version = importlib.metadata.version("uv")
    base_dir = os.path.dirname(os.path.dirname(find_spec("uv").origin))
    new_path = os.path.join(base_dir, f"uv-{version}.data/scripts")
    # set this up because `uv` via @pip does not unpack binaries in discoverable location
    if os.path.exists(new_path):
        os.environ["UV_INSTALL_DIR"] = new_path


@pytest.fixture
def input_requirements_file(tmp_path):
    requirements_path = tmp_path / "requirements.txt"
    requirements_path.write_text(INPUT_REQUIREMENTS_TEXT)
    return requirements_path


@pytest.fixture
def error_input_requirements_file(tmp_path):
    requirements_path = tmp_path / "resolved.txt"
    requirements_path.write_text(ERROR_INPUT_REQUIREMENTS_TEXT)
    return requirements_path


@pytest.fixture
def resolved_requirements_file(tmp_path):
    requirements_path = tmp_path / "resolved.txt"
    requirements_path.write_text(RESOLVED_REQUIREMENTS_TEXT)
    return requirements_path


@dataclass
class FileSplit__TestCase:
    name: str
    file_size: int
    expected_parts: List[UploadPart]


FILE_SPLIT_TEST_CASES = [
    FileSplit__TestCase(
        name="single_file",
        file_size=DEFAULT_UPLOAD_PART_SIZE_MB * 1024 * 1024 - 1,
        expected_parts=[UploadPart(part_number=1, offset=0, part_size=DEFAULT_UPLOAD_PART_SIZE_MB * 1024 * 1024 - 1)],
    ),
    FileSplit__TestCase(
        name="exact_multiple_parts",
        file_size=DEFAULT_UPLOAD_PART_SIZE_MB * 1024 * 1024 * 5,
        expected_parts=[
            UploadPart(
                part_number=i,
                offset=(i - 1) * DEFAULT_UPLOAD_PART_SIZE_MB * 1024 * 1024,
                part_size=DEFAULT_UPLOAD_PART_SIZE_MB * 1024 * 1024,
            )
            for i in range(1, 6)
        ],
    ),
    FileSplit__TestCase(
        name="multiple_parts_with_last_part_smaller",
        file_size=(DEFAULT_UPLOAD_PART_SIZE_MB * 1024 * 1024 * 2) + (DEFAULT_UPLOAD_PART_SIZE_MB * 1024 * 1024 // 2),
        expected_parts=[
            UploadPart(part_number=1, offset=0, part_size=DEFAULT_UPLOAD_PART_SIZE_MB * 1024 * 1024),
            UploadPart(
                part_number=2,
                offset=DEFAULT_UPLOAD_PART_SIZE_MB * 1024 * 1024,
                part_size=DEFAULT_UPLOAD_PART_SIZE_MB * 1024 * 1024,
            ),
            UploadPart(
                part_number=3,
                offset=2 * DEFAULT_UPLOAD_PART_SIZE_MB * 1024 * 1024,
                part_size=DEFAULT_UPLOAD_PART_SIZE_MB * 1024 * 1024 // 2,
            ),
        ],
    ),
    FileSplit__TestCase(
        name="zero_size_file",
        file_size=0,
        expected_parts=[],
    ),
]


@pytest.mark.parametrize("test_case", FILE_SPLIT_TEST_CASES, ids=[tc.name for tc in FILE_SPLIT_TEST_CASES])
def test_get_upload_parts(test_case):
    parts = get_upload_parts(test_case.file_size)
    assert len(parts) == len(test_case.expected_parts)
    for part, expected_part in zip(parts, test_case.expected_parts):
        assert part.part_size == expected_part.part_size
        assert part.part_number == expected_part.part_number
        assert part.offset == expected_part.offset


@pytest.mark.parametrize(
    "name, expected",
    [
        ("env123", True),
        ("env_123", True),
        ("ENV-123", True),
        ("env*123", False),
        ("env?123", False),
        ("env!123", False),
        ("", False),
        ("env 123", False),
        ("env_01234567890123456789001234567890012345678900123456789001234567890", False),
    ],
)
def test_environments(name, expected):
    assert is_valid_environment_name(name) == expected


@pytest.mark.skip("Test requires internet access.")
def test_requirements_resolution_uv(input_requirements_file, tmp_path, uv_binary_setup):
    # tests python dependency resolution (using `uv`)
    resolved_requirements_path, tecton_runtime_version, tecton_rift_version = _run_dependency_resolution(
        input_requirements_file, tmp_path, python_version="3.9", tool="uv"
    )

    assert tecton_runtime_version == "1.0.0"
    assert tecton_rift_version == "1.1.0b6"


@pytest.mark.skip("Test requires internet access. Pex cannot resolve `xformers` because of manylinux compatibility")
def test_requirements_resolution_pex(input_requirements_file, tmp_path):
    # tests python dependency resolution (using `pex`)
    resolved_requirements_path, tecton_runtime_version, tecton_rift_version = _run_dependency_resolution(
        input_requirements_file, tmp_path, python_version="3.9", tool="pex"
    )

    assert tecton_runtime_version == "1.0.0"
    assert tecton_rift_version == "1.1.0b6"


def test_error_requirements_resolution(error_input_requirements_file, tmp_path):
    # exists because requirements does not have any supporting tecton libraries
    with pytest.raises(SystemExit) as e:
        _run_dependency_resolution(error_input_requirements_file, tmp_path, python_version="3.9", tool="uv")
    assert e.type == SystemExit
    assert e.value.code == 1


def test_get_tecton_versions_from_resolved(resolved_requirements_file):
    tecton_pkg_to_version = _get_pkg_to_version(
        resolved_requirements_file, packages=["tecton[rift-materialization]", "tecton-runtime", "deltalake"]
    )
    tecton_runtime_version = tecton_pkg_to_version.get("tecton-runtime")
    tecton_rift_version = tecton_pkg_to_version.get("tecton[rift-materialization]")
    deltalake_version = tecton_pkg_to_version.get("deltalake")

    assert tecton_runtime_version == "1.0.0"
    assert tecton_rift_version == "99.99.99"
    assert deltalake_version == "0.18.2"


@pytest.fixture
def environment_dependencies():
    return EnvironmentDependencies(SAMPLE_LOCK_JSON)


@pytest.mark.parametrize(
    "package_name, expected_version",
    [("attrs", "23.2.0"), ("boto3", "1.34.87"), ("pytest", "8.1.1"), ("nonexistent", None)],
)
def test_get_version(environment_dependencies, package_name, expected_version):
    assert environment_dependencies.get_version(package_name) == expected_version


@pytest.mark.parametrize(
    "package_name, expected_presence", [("attrs", True), ("boto3", True), ("pytest", True), ("nonexistent", False)]
)
def test_is_dependency_present(environment_dependencies, package_name, expected_presence):
    assert environment_dependencies.is_dependency_present(package_name) == expected_presence


@pytest.mark.parametrize(
    "package_name, expected_extras", [("attrs", ["tests-no-zope"]), ("boto3", []), ("pytest", []), ("nonexistent", [])]
)
def test_get_dependency_extras(environment_dependencies, package_name, expected_extras):
    extras = environment_dependencies.get_dependency_extras(package_name)
    assert extras == expected_extras



from google.protobuf import struct_pb2 as _struct_pb2
from tecton_proto.common import python_version__client_pb2 as _python_version__client_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor
GENERAL_PURPOSE_SSD: EbsVolumeType
INSTANCE_FLEET_FOR_INTEGRATION_TESTS_ONLY: AwsAvailability
ON_DEMAND: AwsAvailability
SPOT: AwsAvailability
SPOT_WITH_FALLBACK: AwsAvailability
THROUGHPUT_OPTIMIZED_HDD: EbsVolumeType
UNKNOWN_AWS_AVAILABILITY: AwsAvailability
UNKNOWN_EBS_VOLUME_TYPE: EbsVolumeType

class AutoScale(_message.Message):
    __slots__ = ["max_workers", "min_workers"]
    MAX_WORKERS_FIELD_NUMBER: _ClassVar[int]
    MIN_WORKERS_FIELD_NUMBER: _ClassVar[int]
    max_workers: int
    min_workers: int
    def __init__(self, min_workers: _Optional[int] = ..., max_workers: _Optional[int] = ...) -> None: ...

class AwsAttributes(_message.Message):
    __slots__ = ["availability", "ebs_volume_count", "ebs_volume_size", "ebs_volume_type", "first_on_demand", "instance_profile_arn", "spot_bid_price_percent", "zone_id"]
    AVAILABILITY_FIELD_NUMBER: _ClassVar[int]
    EBS_VOLUME_COUNT_FIELD_NUMBER: _ClassVar[int]
    EBS_VOLUME_SIZE_FIELD_NUMBER: _ClassVar[int]
    EBS_VOLUME_TYPE_FIELD_NUMBER: _ClassVar[int]
    FIRST_ON_DEMAND_FIELD_NUMBER: _ClassVar[int]
    INSTANCE_PROFILE_ARN_FIELD_NUMBER: _ClassVar[int]
    SPOT_BID_PRICE_PERCENT_FIELD_NUMBER: _ClassVar[int]
    ZONE_ID_FIELD_NUMBER: _ClassVar[int]
    availability: AwsAvailability
    ebs_volume_count: int
    ebs_volume_size: int
    ebs_volume_type: EbsVolumeType
    first_on_demand: int
    instance_profile_arn: str
    spot_bid_price_percent: int
    zone_id: str
    def __init__(self, first_on_demand: _Optional[int] = ..., availability: _Optional[_Union[AwsAvailability, str]] = ..., zone_id: _Optional[str] = ..., spot_bid_price_percent: _Optional[int] = ..., instance_profile_arn: _Optional[str] = ..., ebs_volume_type: _Optional[_Union[EbsVolumeType, str]] = ..., ebs_volume_count: _Optional[int] = ..., ebs_volume_size: _Optional[int] = ...) -> None: ...

class ClusterInfo(_message.Message):
    __slots__ = ["final_json", "new_cluster", "warnings"]
    FINAL_JSON_FIELD_NUMBER: _ClassVar[int]
    NEW_CLUSTER_FIELD_NUMBER: _ClassVar[int]
    WARNINGS_FIELD_NUMBER: _ClassVar[int]
    final_json: str
    new_cluster: NewCluster
    warnings: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, new_cluster: _Optional[_Union[NewCluster, _Mapping]] = ..., final_json: _Optional[str] = ..., warnings: _Optional[_Iterable[str]] = ...) -> None: ...

class ClusterLogConf(_message.Message):
    __slots__ = ["s3"]
    S3_FIELD_NUMBER: _ClassVar[int]
    s3: S3StorageInfo
    def __init__(self, s3: _Optional[_Union[S3StorageInfo, _Mapping]] = ...) -> None: ...

class ClusterTag(_message.Message):
    __slots__ = ["key", "value"]
    KEY_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    key: str
    value: str
    def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...

class DbfsStorageInfo(_message.Message):
    __slots__ = ["destination"]
    DESTINATION_FIELD_NUMBER: _ClassVar[int]
    destination: str
    def __init__(self, destination: _Optional[str] = ...) -> None: ...

class ExistingCluster(_message.Message):
    __slots__ = ["cluster_id"]
    CLUSTER_ID_FIELD_NUMBER: _ClassVar[int]
    cluster_id: str
    def __init__(self, cluster_id: _Optional[str] = ...) -> None: ...

class GCPAttributes(_message.Message):
    __slots__ = ["availability", "boot_disk_size", "google_service_account", "use_preemptible_executors", "zone_id"]
    AVAILABILITY_FIELD_NUMBER: _ClassVar[int]
    BOOT_DISK_SIZE_FIELD_NUMBER: _ClassVar[int]
    GOOGLE_SERVICE_ACCOUNT_FIELD_NUMBER: _ClassVar[int]
    USE_PREEMPTIBLE_EXECUTORS_FIELD_NUMBER: _ClassVar[int]
    ZONE_ID_FIELD_NUMBER: _ClassVar[int]
    availability: str
    boot_disk_size: int
    google_service_account: str
    use_preemptible_executors: bool
    zone_id: str
    def __init__(self, use_preemptible_executors: bool = ..., google_service_account: _Optional[str] = ..., boot_disk_size: _Optional[int] = ..., availability: _Optional[str] = ..., zone_id: _Optional[str] = ...) -> None: ...

class LocalStorageInfo(_message.Message):
    __slots__ = ["path"]
    PATH_FIELD_NUMBER: _ClassVar[int]
    path: str
    def __init__(self, path: _Optional[str] = ...) -> None: ...

class NewCluster(_message.Message):
    __slots__ = ["apply_policy_default_values", "autoscale", "aws_attributes", "cluster_log_conf", "cluster_name", "custom_tags", "data_security_mode", "driver_node_type_id", "enable_elastic_disk", "gcp_attributes", "init_scripts", "instance_pool_id", "json_cluster_config", "node_type_id", "num_workers", "policy_id", "python_version", "root_volume_size_in_gb", "single_user_name", "spark_conf", "spark_env_vars", "spark_version", "terminateOnComplete"]
    class SparkConfEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    class SparkEnvVarsEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    APPLY_POLICY_DEFAULT_VALUES_FIELD_NUMBER: _ClassVar[int]
    AUTOSCALE_FIELD_NUMBER: _ClassVar[int]
    AWS_ATTRIBUTES_FIELD_NUMBER: _ClassVar[int]
    CLUSTER_LOG_CONF_FIELD_NUMBER: _ClassVar[int]
    CLUSTER_NAME_FIELD_NUMBER: _ClassVar[int]
    CUSTOM_TAGS_FIELD_NUMBER: _ClassVar[int]
    DATA_SECURITY_MODE_FIELD_NUMBER: _ClassVar[int]
    DRIVER_NODE_TYPE_ID_FIELD_NUMBER: _ClassVar[int]
    ENABLE_ELASTIC_DISK_FIELD_NUMBER: _ClassVar[int]
    GCP_ATTRIBUTES_FIELD_NUMBER: _ClassVar[int]
    INIT_SCRIPTS_FIELD_NUMBER: _ClassVar[int]
    INSTANCE_POOL_ID_FIELD_NUMBER: _ClassVar[int]
    JSON_CLUSTER_CONFIG_FIELD_NUMBER: _ClassVar[int]
    NODE_TYPE_ID_FIELD_NUMBER: _ClassVar[int]
    NUM_WORKERS_FIELD_NUMBER: _ClassVar[int]
    POLICY_ID_FIELD_NUMBER: _ClassVar[int]
    PYTHON_VERSION_FIELD_NUMBER: _ClassVar[int]
    ROOT_VOLUME_SIZE_IN_GB_FIELD_NUMBER: _ClassVar[int]
    SINGLE_USER_NAME_FIELD_NUMBER: _ClassVar[int]
    SPARK_CONF_FIELD_NUMBER: _ClassVar[int]
    SPARK_ENV_VARS_FIELD_NUMBER: _ClassVar[int]
    SPARK_VERSION_FIELD_NUMBER: _ClassVar[int]
    TERMINATEONCOMPLETE_FIELD_NUMBER: _ClassVar[int]
    apply_policy_default_values: bool
    autoscale: AutoScale
    aws_attributes: AwsAttributes
    cluster_log_conf: ResourceLocation
    cluster_name: str
    custom_tags: _containers.RepeatedCompositeFieldContainer[ClusterTag]
    data_security_mode: str
    driver_node_type_id: str
    enable_elastic_disk: bool
    gcp_attributes: GCPAttributes
    init_scripts: _containers.RepeatedCompositeFieldContainer[ResourceLocation]
    instance_pool_id: str
    json_cluster_config: _struct_pb2.Struct
    node_type_id: str
    num_workers: int
    policy_id: str
    python_version: _python_version__client_pb2.PythonVersion
    root_volume_size_in_gb: int
    single_user_name: str
    spark_conf: _containers.ScalarMap[str, str]
    spark_env_vars: _containers.ScalarMap[str, str]
    spark_version: str
    terminateOnComplete: bool
    def __init__(self, num_workers: _Optional[int] = ..., autoscale: _Optional[_Union[AutoScale, _Mapping]] = ..., cluster_name: _Optional[str] = ..., spark_version: _Optional[str] = ..., spark_conf: _Optional[_Mapping[str, str]] = ..., aws_attributes: _Optional[_Union[AwsAttributes, _Mapping]] = ..., node_type_id: _Optional[str] = ..., enable_elastic_disk: bool = ..., init_scripts: _Optional[_Iterable[_Union[ResourceLocation, _Mapping]]] = ..., cluster_log_conf: _Optional[_Union[ResourceLocation, _Mapping]] = ..., custom_tags: _Optional[_Iterable[_Union[ClusterTag, _Mapping]]] = ..., terminateOnComplete: bool = ..., spark_env_vars: _Optional[_Mapping[str, str]] = ..., gcp_attributes: _Optional[_Union[GCPAttributes, _Mapping]] = ..., json_cluster_config: _Optional[_Union[_struct_pb2.Struct, _Mapping]] = ..., policy_id: _Optional[str] = ..., apply_policy_default_values: bool = ..., driver_node_type_id: _Optional[str] = ..., root_volume_size_in_gb: _Optional[int] = ..., python_version: _Optional[_Union[_python_version__client_pb2.PythonVersion, str]] = ..., data_security_mode: _Optional[str] = ..., single_user_name: _Optional[str] = ..., instance_pool_id: _Optional[str] = ...) -> None: ...

class ResourceLocation(_message.Message):
    __slots__ = ["dbfs", "local", "s3", "workspace"]
    DBFS_FIELD_NUMBER: _ClassVar[int]
    LOCAL_FIELD_NUMBER: _ClassVar[int]
    S3_FIELD_NUMBER: _ClassVar[int]
    WORKSPACE_FIELD_NUMBER: _ClassVar[int]
    dbfs: DbfsStorageInfo
    local: LocalStorageInfo
    s3: S3StorageInfo
    workspace: WorkspaceStorageInfo
    def __init__(self, s3: _Optional[_Union[S3StorageInfo, _Mapping]] = ..., dbfs: _Optional[_Union[DbfsStorageInfo, _Mapping]] = ..., workspace: _Optional[_Union[WorkspaceStorageInfo, _Mapping]] = ..., local: _Optional[_Union[LocalStorageInfo, _Mapping]] = ...) -> None: ...

class S3StorageInfo(_message.Message):
    __slots__ = ["destination", "region"]
    DESTINATION_FIELD_NUMBER: _ClassVar[int]
    REGION_FIELD_NUMBER: _ClassVar[int]
    destination: str
    region: str
    def __init__(self, destination: _Optional[str] = ..., region: _Optional[str] = ...) -> None: ...

class WorkspaceStorageInfo(_message.Message):
    __slots__ = ["destination"]
    DESTINATION_FIELD_NUMBER: _ClassVar[int]
    destination: str
    def __init__(self, destination: _Optional[str] = ...) -> None: ...

class AwsAvailability(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class EbsVolumeType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []



from google.protobuf import duration_pb2 as _duration_pb2
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

CLIENT_LOG_MESSAGE_TYPE_METHOD_ENTRY: LogMessageType
CLIENT_LOG_MESSAGE_TYPE_METHOD_RETURN: LogMessageType
CLIENT_LOG_MESSAGE_TYPE_UNKNOWN: LogMessageType
DESCRIPTOR: _descriptor.FileDescriptor

class ErrorLog(_message.Message):
    __slots__ = ["cause", "message", "stacktrace"]
    CAUSE_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    STACKTRACE_FIELD_NUMBER: _ClassVar[int]
    cause: ErrorLog
    message: str
    stacktrace: str
    def __init__(self, message: _Optional[str] = ..., stacktrace: _Optional[str] = ..., cause: _Optional[_Union[ErrorLog, _Mapping]] = ...) -> None: ...

class LoggedValue(_message.Message):
    __slots__ = ["name", "value"]
    NAME_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    name: str
    value: str
    def __init__(self, name: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...

class SDKMethodInvocation(_message.Message):
    __slots__ = ["class_name", "error", "execution_time", "is_local_fco", "log_level", "method_name", "params_or_return_values", "python_version", "sdk_version", "time", "trace_id", "type", "user_id", "workspace"]
    CLASS_NAME_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    EXECUTION_TIME_FIELD_NUMBER: _ClassVar[int]
    IS_LOCAL_FCO_FIELD_NUMBER: _ClassVar[int]
    LOG_LEVEL_FIELD_NUMBER: _ClassVar[int]
    METHOD_NAME_FIELD_NUMBER: _ClassVar[int]
    PARAMS_OR_RETURN_VALUES_FIELD_NUMBER: _ClassVar[int]
    PYTHON_VERSION_FIELD_NUMBER: _ClassVar[int]
    SDK_VERSION_FIELD_NUMBER: _ClassVar[int]
    TIME_FIELD_NUMBER: _ClassVar[int]
    TRACE_ID_FIELD_NUMBER: _ClassVar[int]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    USER_ID_FIELD_NUMBER: _ClassVar[int]
    WORKSPACE_FIELD_NUMBER: _ClassVar[int]
    class_name: str
    error: ErrorLog
    execution_time: _duration_pb2.Duration
    is_local_fco: bool
    log_level: str
    method_name: str
    params_or_return_values: _containers.RepeatedCompositeFieldContainer[LoggedValue]
    python_version: str
    sdk_version: str
    time: _timestamp_pb2.Timestamp
    trace_id: str
    type: LogMessageType
    user_id: str
    workspace: str
    def __init__(self, time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., user_id: _Optional[str] = ..., trace_id: _Optional[str] = ..., log_level: _Optional[str] = ..., type: _Optional[_Union[LogMessageType, str]] = ..., class_name: _Optional[str] = ..., method_name: _Optional[str] = ..., execution_time: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ..., params_or_return_values: _Optional[_Iterable[_Union[LoggedValue, _Mapping]]] = ..., error: _Optional[_Union[ErrorLog, _Mapping]] = ..., workspace: _Optional[str] = ..., sdk_version: _Optional[str] = ..., python_version: _Optional[str] = ..., is_local_fco: bool = ...) -> None: ...

class LogMessageType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []



from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class ListSecretScopesResponse(_message.Message):
    __slots__ = ["scopes"]
    SCOPES_FIELD_NUMBER: _ClassVar[int]
    scopes: _containers.RepeatedCompositeFieldContainer[SecretScopeInfo]
    def __init__(self, scopes: _Optional[_Iterable[_Union[SecretScopeInfo, _Mapping]]] = ...) -> None: ...

class PutRequest(_message.Message):
    __slots__ = ["contents", "overwrite", "path"]
    CONTENTS_FIELD_NUMBER: _ClassVar[int]
    OVERWRITE_FIELD_NUMBER: _ClassVar[int]
    PATH_FIELD_NUMBER: _ClassVar[int]
    contents: bytes
    overwrite: bool
    path: str
    def __init__(self, path: _Optional[str] = ..., contents: _Optional[bytes] = ..., overwrite: bool = ...) -> None: ...

class ScopeCreateRequest(_message.Message):
    __slots__ = ["scope"]
    SCOPE_FIELD_NUMBER: _ClassVar[int]
    scope: str
    def __init__(self, scope: _Optional[str] = ...) -> None: ...

class SecretAclPutRequest(_message.Message):
    __slots__ = ["permission", "principal", "scope"]
    PERMISSION_FIELD_NUMBER: _ClassVar[int]
    PRINCIPAL_FIELD_NUMBER: _ClassVar[int]
    SCOPE_FIELD_NUMBER: _ClassVar[int]
    permission: str
    principal: str
    scope: str
    def __init__(self, scope: _Optional[str] = ..., principal: _Optional[str] = ..., permission: _Optional[str] = ...) -> None: ...

class SecretPutRequest(_message.Message):
    __slots__ = ["key", "scope", "string_value"]
    KEY_FIELD_NUMBER: _ClassVar[int]
    SCOPE_FIELD_NUMBER: _ClassVar[int]
    STRING_VALUE_FIELD_NUMBER: _ClassVar[int]
    key: str
    scope: str
    string_value: str
    def __init__(self, scope: _Optional[str] = ..., key: _Optional[str] = ..., string_value: _Optional[str] = ...) -> None: ...

class SecretScopeInfo(_message.Message):
    __slots__ = ["backend_type", "name"]
    BACKEND_TYPE_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    backend_type: str
    name: str
    def __init__(self, name: _Optional[str] = ..., backend_type: _Optional[str] = ...) -> None: ...



import sys
from datetime import datetime
from typing import Any
from typing import Dict
from typing import Union


if sys.version_info >= (3, 8):
    from typing import Protocol
else:
    from typing_extensions import Protocol

import pandas

from tecton_core import specs
from tecton_core.id_helper import IdHelper
from tecton_proto.args.pipeline__client_pb2 import PipelineNode
from tecton_proto.common.id__client_pb2 import Id as TectonId


NodeValueType = Union[str, int, float, bool, datetime, None, Dict[str, Any], pandas.DataFrame, pandas.Series]


class FeaturePipeline(Protocol):
    def _node_to_value(self, pipeline_node: PipelineNode) -> NodeValueType:
        """Evaluates a node of this Feature Pipeline to a value"""
        ...

    def _transformation_node_to_value(self, pipeline_node: PipelineNode) -> Union[Dict[str, Any], pandas.DataFrame]:
        """Evaluates the TransformationNode of this Feature Pipeline to a value"""
        ...

    def __init__(self):
        self._id_to_transformation = {}
        self._pipeline = None

    def get_dataframe(self):
        """Run this Feature Pipeline to generate a result DataFrame"""
        return self._node_to_value(self._pipeline.root)

    def to_dataframe(self):
        return self.get_dataframe()

    def get_transformation_by_id(self, id: TectonId) -> specs.TransformationSpec:
        return self._id_to_transformation[IdHelper.to_string(id)]



from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class AwsCredentials(_message.Message):
    __slots__ = ["access_key_id", "expiration", "secret_access_key", "session_token"]
    ACCESS_KEY_ID_FIELD_NUMBER: _ClassVar[int]
    EXPIRATION_FIELD_NUMBER: _ClassVar[int]
    SECRET_ACCESS_KEY_FIELD_NUMBER: _ClassVar[int]
    SESSION_TOKEN_FIELD_NUMBER: _ClassVar[int]
    access_key_id: str
    expiration: _timestamp_pb2.Timestamp
    secret_access_key: str
    session_token: str
    def __init__(self, access_key_id: _Optional[str] = ..., secret_access_key: _Optional[str] = ..., session_token: _Optional[str] = ..., expiration: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...) -> None: ...

class AwsIamRole(_message.Message):
    __slots__ = ["external_id", "intermediate_role", "role_arn"]
    EXTERNAL_ID_FIELD_NUMBER: _ClassVar[int]
    INTERMEDIATE_ROLE_FIELD_NUMBER: _ClassVar[int]
    ROLE_ARN_FIELD_NUMBER: _ClassVar[int]
    external_id: str
    intermediate_role: AwsIamRole
    role_arn: str
    def __init__(self, role_arn: _Optional[str] = ..., intermediate_role: _Optional[_Union[AwsIamRole, _Mapping]] = ..., external_id: _Optional[str] = ...) -> None: ...



from google.protobuf import duration_pb2 as _duration_pb2
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor
INITIAL_STREAM_POSITION_LATEST: InitialStreamPosition
INITIAL_STREAM_POSITION_TRIM_HORIZON: InitialStreamPosition
INITIAL_STREAM_POSITION_UNSPECIFIED: InitialStreamPosition

class BatchConfig(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...

class StreamConfig(_message.Message):
    __slots__ = ["initial_stream_position", "watermark_delay_threshold"]
    INITIAL_STREAM_POSITION_FIELD_NUMBER: _ClassVar[int]
    WATERMARK_DELAY_THRESHOLD_FIELD_NUMBER: _ClassVar[int]
    initial_stream_position: InitialStreamPosition
    watermark_delay_threshold: _duration_pb2.Duration
    def __init__(self, initial_stream_position: _Optional[_Union[InitialStreamPosition, str]] = ..., watermark_delay_threshold: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ...) -> None: ...

class InitialStreamPosition(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []



import datetime
import re
from typing import TYPE_CHECKING
from typing import Optional
from typing import TypeVar
from typing import Union

import pandas
import pytimeparse
from google.protobuf import duration_pb2
from google.protobuf import timestamp_pb2
from pypika.terms import Term
from typeguard import typechecked

import tecton_core.tecton_pendulum as pendulum
from tecton_core.errors import TectonValidationError
from tecton_proto.data import feature_store__client_pb2 as feature_store_pb2


if TYPE_CHECKING:
    from tecton_core.feature_definition_wrapper import FeatureDefinitionWrapper


# 10,000 years which is less than but still close to the max duration proto amount, and a digestable number of years to
# a user
_MAX_PROTO_DURATION_SECONDS = 10_000 * 365 * 24 * 60 * 60

_TIME_RANGE_REGEX = r"^(?:(\d+)d)?(?:(\d+)h)?(?:(\d+)m)?(?:(\d+)s)?$"

TERM_OR_INT = TypeVar("TERM_OR_INT", int, Term)


def timedelta_to_duration(td: datetime.timedelta) -> pendulum.Duration:
    return pendulum.duration(days=td.days, seconds=td.seconds, microseconds=td.microseconds)


def proto_to_duration(proto_duration: duration_pb2.Duration) -> pendulum.Duration:
    return timedelta_to_duration(proto_duration.ToTimedelta())


def timedelta_to_proto(td: Optional[datetime.timedelta]) -> Optional[duration_pb2.Duration]:
    if td is None:
        return None

    if td.total_seconds() > _MAX_PROTO_DURATION_SECONDS:
        msg = "'Timedelta value must be lower than max value of 10000 years'"
        raise TectonValidationError(msg)

    proto = duration_pb2.Duration()
    proto.FromTimedelta(td)
    return proto


def to_positive_proto(proto_duration: duration_pb2.Duration) -> duration_pb2.Duration:
    return duration_pb2.Duration(seconds=abs(duration_pb2.Duration.seconds), nanos=abs(duration_pb2.Duration.nanos))


def datetime_to_proto(dt: Optional[datetime.datetime]) -> Optional[timestamp_pb2.Timestamp]:
    if dt is None:
        return None
    proto = timestamp_pb2.Timestamp()
    proto.FromDatetime(dt)
    return proto


@typechecked
def to_human_readable_str(duration: Union[duration_pb2.Duration, datetime.timedelta]) -> str:
    """
    Replicate the backend logic that converts durations to feature strings, e.g. `column_sum_1h30m_30m`.

    Backend logic: https://github.com/tecton-ai/tecton/blob/e3c04e28066c21fa4ed9b459ece458c084766b6e/java/com/tecton/common/utils/TimeUtils.kt#L80
    """

    if isinstance(duration, duration_pb2.Duration):
        duration = proto_to_duration(duration)
    elif isinstance(duration, datetime.timedelta):
        duration = timedelta_to_duration(duration)
    else:
        msg = f"Invalid input duration type: {type(duration)}"
        raise TypeError(msg)

    duration_str_list = []
    if duration.days > 0:
        duration_str_list.append(f"{duration.days}d")
    if duration.hours > 0:
        duration_str_list.append(f"{duration.hours}h")
    if duration.minutes > 0:
        duration_str_list.append(f"{duration.minutes}m")
    if duration.remaining_seconds > 0:
        duration_str_list.append(f"{duration.remaining_seconds}s")

    return "0s" if len(duration_str_list) == 0 else "".join(duration_str_list)


@typechecked
def parse_to_timedelta(time_string: str) -> datetime.timedelta:
    """Parse a string with format '{days}d{hours}h{minutes}m{seconds}s' into a timedelta object.

    Each time component is optional, but at least one value must be provided, e.g. "" is invalid but "0s" is valid.
    """
    if time_string == "":
        msg = "Cannot use an empty string for a time range. Use '0d', '0s', etc. instead."
        raise ValueError(msg)

    match = re.match(_TIME_RANGE_REGEX, time_string)
    if not match:
        msg = f"The time range string '{time_string}' is not in the correct format. Should be of the form '[days]d[hours]h[minutes]m[seconds]s', e.g. `1d12h`."
        raise ValueError(msg)

    # Extract time components from the matched groups, substituting 0 where no component is found
    days, hours, minutes, seconds = [int(g) if g else 0 for g in match.groups()]

    # Create a timedelta object with the extracted time components
    return datetime.timedelta(days=days, hours=hours, minutes=minutes, seconds=seconds)


def assert_is_round_seconds(d: pendulum.Duration) -> pendulum.Duration:
    if d % pendulum.Duration(seconds=1):
        msg = f"{d.in_words()} is not a round number of seconds"
        raise ValueError(msg)
    return d


def strict_pytimeparse(time_str: str) -> Union[int, float]:
    parsed = pytimeparse.parse(time_str)
    if parsed is None:
        msg = f'Could not parse time string "{time_str}"'
        raise TectonValidationError(msg)
    else:
        return parsed


def nanos_to_seconds(nanos: int) -> float:
    """
    :param nanos: Nanoseconds
    :return: Converts nanoseconds to seconds
    """
    return nanos / float(1e9)


def seconds_to_nanos(seconds: float) -> int:
    """
    :param seconds: Seconds
    :return: Converts seconds to nanoseconds
    """
    return int(seconds * 1e9)


def convert_duration_to_seconds(duration: int, version: int) -> int:
    """Convert a duration to seconds according to the version.

    If V0, the duration is already in seconds; if V1, the duration must be converted from nanos to seconds.
    """
    if version == feature_store_pb2.FEATURE_STORE_FORMAT_VERSION_DEFAULT:
        return duration
    else:
        return int(nanos_to_seconds(duration))


def convert_timedelta_for_version(duration: datetime.timedelta, version: int) -> int:
    """
    Convert pendulum duration according to version
    VO -> Return Seconds
    V1 -> Return Nanoseconds
    :param duration: Pendulum Duration
    :param version: Feature Store Format Version
    :return:
    """
    assert duration.microseconds == 0
    interval = duration.total_seconds()
    if version == feature_store_pb2.FEATURE_STORE_FORMAT_VERSION_DEFAULT:
        return int(interval)
    else:
        return seconds_to_nanos(interval)


def convert_epoch_to_datetime(epoch: int, version: int) -> pendulum.datetime:
    """
    Converts an epoch to a datetime.
    """
    if version == feature_store_pb2.FeatureStoreFormatVersion.FEATURE_STORE_FORMAT_VERSION_DEFAULT:
        return pendulum.from_timestamp(epoch)
    else:
        return pendulum.from_timestamp(epoch / 1e9)


def convert_timestamp_for_version(timestamp: datetime.datetime, version: int) -> int:
    if version == feature_store_pb2.FeatureStoreFormatVersion.FEATURE_STORE_FORMAT_VERSION_DEFAULT:
        return int(timestamp.timestamp())
    else:
        return seconds_to_nanos(timestamp.timestamp())


def convert_proto_duration_for_version(duration: duration_pb2.Duration, version: int) -> int:
    return convert_timedelta_for_version(duration.ToTimedelta(), version)


def align_time_downwards(time: datetime.datetime, alignment: datetime.timedelta) -> datetime.datetime:
    excess_seconds = time.timestamp() % alignment.total_seconds()
    return datetime.datetime.utcfromtimestamp(time.timestamp() - excess_seconds)


def align_epoch_downwards(int_timestamp_col: int, window_size: int) -> int:
    return int_timestamp_col - (int_timestamp_col % window_size)


def align_time_upwards(time: datetime.datetime, alignment: datetime.timedelta) -> datetime.datetime:
    excess_seconds = time.timestamp() % alignment.total_seconds()
    offset = 0 if excess_seconds == 0 else alignment.total_seconds() - excess_seconds
    return datetime.datetime.utcfromtimestamp(time.timestamp() + offset)


def get_timezone_aware_datetime(time: Optional[datetime.datetime]) -> Optional[datetime.datetime]:
    """
    Check and get timezone aware datetime
    :param time: Time to convert to timezone aware
    """
    if time is None:
        return None
    if time.tzinfo is None:
        return time.replace(tzinfo=pendulum.timezone("UTC"))
    else:
        return time


def convert_pandas_df_for_snowflake_upload(pandas_df: pandas.DataFrame) -> None:
    """
    Converts a pandas dataframe's timestamp columns to ensure expected behavior when uploading to Snowflake.

    Snowflake's `write_pandas` logic only writes TIMESTAMP_NTZ timestamps correctly if we
    pretend the input timestamps are UTC. Providing another time zone (e.g. 11PM PST) will convert
    this to UTC time (6AM UTC), and then drop the timezone info + load as a TIMESTAMP_NTZ column (6AM).
    """
    # TODO(TEC-16678): Revert this once a fix is upstreamed to Snowflake
    for col in pandas_df.columns:
        if pandas.core.dtypes.common.is_datetime64_dtype(pandas_df[col]):
            pandas_df[col] = pandas_df[col].dt.tz_localize(None)
            pandas_df[col] = pandas_df[col].dt.tz_localize(pendulum.timezone("UTC"))


def convert_to_effective_timestamp(
    timestamp_col: TERM_OR_INT, batch_schedule_seconds: int, data_delay_seconds: int
) -> TERM_OR_INT:
    """
    Converts a timestamp col term or integer and converts it into an effective timestamp by aligning on the interval
    and adding the interval and data delay.

    The effective timestamp represents the time when the feature is available in the online store to be used.
    """
    return timestamp_col - (timestamp_col % batch_schedule_seconds) + batch_schedule_seconds + data_delay_seconds


def get_nearest_anchor_time(
    timestamp: pendulum.DateTime,
    max_source_data_delay: pendulum.Duration,
    batch_materialization_schedule: pendulum.Duration,
    min_scheduling_interval: pendulum.Duration,
) -> pendulum.DateTime:
    """
    Returns the nearest anchor time (start of tile), such that the feature values as of this time would have the
    effective timestamp `timestamp`
    """
    timestamp = timestamp - max_source_data_delay - batch_materialization_schedule

    if min_scheduling_interval.total_seconds() > 0:
        timestamp = align_time_downwards(timestamp, min_scheduling_interval)

    return pendulum.instance(timestamp)


def align_left(timestamp: pendulum.DateTime, min_scheduling_interval: pendulum.Duration) -> pendulum.DateTime:
    # min scheduling interval may be zero for continuous
    if min_scheduling_interval and min_scheduling_interval.total_seconds() > 0:
        timestamp = align_time_downwards(timestamp, min_scheduling_interval)
    return pendulum.instance(timestamp)


def align_right(timestamp: pendulum.DateTime, min_scheduling_interval: pendulum.Duration) -> pendulum.DateTime:
    """Aligns timestamp to the end of the scheduling interval. If it's already aligned,
    adds scheduling interval's length.

    This function does not take allowed lateness into account.

    :param timestamp: The timestamp in python datetime format
    :return: The timestamp of the greatest aligned time <= timestamp, in seconds.
    """
    aligned_left = align_left(timestamp, min_scheduling_interval)

    # input timestamp is already aligned
    if aligned_left == timestamp:
        return timestamp

    return aligned_left + min_scheduling_interval


def temporal_fv_get_feature_data_time_limits(
    fd: "FeatureDefinitionWrapper",
    spine_time_limits: pendulum.Period,
    max_lookback: Optional[datetime.timedelta],
) -> pendulum.Period:
    """Feature data time limits for a non aggregate feature view.

    Accounts for:
      * serving ttl
      * batch schedule
      * data delay

    Does NOT account for:
      * batch vs streaming

    NOTE: for historical reasons, this is not an exact filter but a more permissive one.
    """

    start_time = spine_time_limits.start
    end_time = spine_time_limits.end

    if max_lookback:
        start_time = start_time - max_lookback
    elif fd.serving_ttl:
        # Subtract by `serving_ttl` to accommodate for enough feature data for the feature expiration
        # We need to account for the data delay + ttl when determining the feature data time limits from the spine.
        start_time = start_time - fd.serving_ttl - fd.max_source_data_delay
    else:
        start_time = pendulum.from_timestamp(0)

    # Respect feature_start_time if it's set.
    if fd.feature_start_timestamp:
        start_time = max(start_time, fd.feature_start_timestamp)

    start_time = align_left(start_time, fd.min_scheduling_interval)

    # Since feature data time interval is open on the right, we need to always strictly align right so that
    # with `batch_schedule = 1h`, time end `04:00:00` will be aligned to `05:00:00`.
    # NOTE: This may be more permissive than 'max_source_data_delay' would allow,
    # but that's okay from a correctness perspective since our as-of join
    # should account for this.
    end_time = align_right(end_time, fd.min_scheduling_interval)

    # It is possible for the start time to be greater than the end time, for example if the feature start time is later
    # than the spine end time. If this happens, we simply set the start time to equal the end time.
    if start_time > end_time:
        start_time = end_time
    return pendulum.Period(start_time, end_time)


def get_feature_data_time_limits(
    fd: "FeatureDefinitionWrapper",
    spine_time_limits: pendulum.Period,
) -> pendulum.Period:
    """Returns the feature data time limits based on the spine time limits, taking aggregations into account.

    To get the raw data time limits, you need to use additional information that is on your FilteredSource or other data source inputs.

    Args:
        fd: The feature view for which to compute feature data time limits.
        spine_time_limits: The time limits of the spine; the start time is inclusive and the end time is exclusive.
    """
    if fd.is_feature_table:
        return _feature_table_get_feature_data_time_limits(fd, spine_time_limits)
    elif fd.is_temporal:
        return temporal_fv_get_feature_data_time_limits(fd, spine_time_limits, None)
    elif fd.is_temporal_aggregate:
        return _temporal_agg_fv_get_feature_data_time_limits(fd, spine_time_limits)

    # Should never happen!
    msg = "Feature definition must be a feature table, temporal FV, or temporal agg FV"
    raise Exception(msg)


def _feature_table_get_feature_data_time_limits(
    fd: "FeatureDefinitionWrapper",
    spine_time_limits: pendulum.Period,
) -> pendulum.Period:
    """Feature data time limits for a feature table.

    This is `[spine_start_time - ttl, spine_end_time)`
    """
    start_time = spine_time_limits.start
    end_time = spine_time_limits.end

    # Subtract by `serving_ttl` to accommodate for enough feature data for the feature expiration
    if fd.serving_ttl:
        start_time = start_time - fd.serving_ttl
    else:
        start_time = pendulum.from_timestamp(0)

    return end_time - start_time


# TODO(TEC-17311): replace pendulum with stanard datetime + support 'unbounded'/None times
def _temporal_agg_fv_get_feature_data_time_limits(
    fd: "FeatureDefinitionWrapper",
    spine_time_limits: pendulum.Period,
) -> pendulum.Period:
    """Feature data time limits for an aggregate feature view.

    Accounts for:
      * aggregation tile interval
      * data delay
      * batch sawtooth tile size

    Does NOT account for:
      * batch vs streaming

    NOTE: for historical reasons, this is not an exact filter but a more permissive one.
    """
    start_time = spine_time_limits.start
    end_time = spine_time_limits.end

    # Respect feature_start_time if it's set.
    if fd.feature_start_timestamp:
        start_time = max(start_time, fd.feature_start_timestamp)

    start_time = align_left(start_time - fd.max_source_data_delay, fd.min_scheduling_interval)
    if fd.get_max_batch_sawtooth_tile_size():
        start_time = align_left(start_time, fd.get_max_batch_sawtooth_tile_size())

    # Account for final aggregation needing aggregation window prior to earliest timestamp
    if fd.has_lifetime_aggregate:
        start_time = pendulum.from_timestamp(0)
    else:
        earliest_window_start = fd.earliest_window_start
        start_time = start_time + timedelta_to_duration(earliest_window_start)

    if fd.materialization_start_timestamp:
        start_time = max(start_time, fd.materialization_start_timestamp)

    # Since feature data time interval is open on the right, we need to always strictly align right so that
    # with `batch_schedule = 1h`, time end `04:00:00` will be aligned to `05:00:00`.
    # NOTE: This may be more permissive than 'max_source_data_delay' would allow,
    # but that's okay from a correctness perspective since our as-of join
    # should account for this.
    end_time = align_right(end_time, fd.min_scheduling_interval)
    if fd.get_max_batch_sawtooth_tile_size():
        end_time = align_right(end_time, fd.get_max_batch_sawtooth_tile_size())

    # It is possible for the start time to be greater than the end time, for example if the feature start time is later
    # than the spine end time. If this happens, we simply set the start time to equal the end time.
    if start_time > end_time:
        start_time = end_time
    return pendulum.Period(start_time, end_time)


def get_feature_data_time_limits_for_lookback(
    fd: "FeatureDefinitionWrapper",
    spine_time_limits: pendulum.Period,
    max_lookback_duration: pendulum.Duration,
) -> pendulum.Period:
    """Aligns the spine time limits on the left to the lookback duration, and on the right to the nearest batch schedule/aggregation interval.

    Accounts for:
      * feature start time
      * data delay
      * materialization start time
      * look back duration parameter
      * batch schedule
      * batch sawtooth tile size

    Does NOT account for:
      * ttl
      * aggregation tile interval
      * aggregation window size
    """
    start_time = spine_time_limits.start
    end_time = spine_time_limits.end

    # Respect feature_start_time if it's set.
    if fd.feature_start_timestamp:
        start_time = max(start_time, fd.feature_start_timestamp)

    start_time = align_left(start_time - fd.max_source_data_delay - max_lookback_duration, fd.min_scheduling_interval)
    if fd.get_max_batch_sawtooth_tile_size():
        start_time = align_left(start_time, fd.get_max_batch_sawtooth_tile_size())

    if fd.materialization_start_timestamp:
        start_time = max(start_time, fd.materialization_start_timestamp)

    end_time = align_right(end_time, fd.min_scheduling_interval)
    if fd.get_max_batch_sawtooth_tile_size():
        end_time = align_right(end_time, fd.get_max_batch_sawtooth_tile_size())
    # It is possible for the start time to be greater than the end time, for example if the feature start time is later
    # than the spine end time. If this happens, we simply set the start time to equal the end time.
    if start_time > end_time:
        start_time = end_time
    return pendulum.Period(start_time, end_time)



from tecton_proto.auth import resource__client_pb2 as _resource__client_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor
ROLE_ASSIGNMENT_TYPE_DIRECT: RoleAssignmentType
ROLE_ASSIGNMENT_TYPE_FROM_PRINCIPAL_GROUP: RoleAssignmentType
ROLE_ASSIGNMENT_TYPE_UNSPECIFIED: RoleAssignmentType

class ResourceAndRoleAssignments(_message.Message):
    __slots__ = ["resource_id", "resource_type", "roles", "roles_granted"]
    RESOURCE_ID_FIELD_NUMBER: _ClassVar[int]
    RESOURCE_TYPE_FIELD_NUMBER: _ClassVar[int]
    ROLES_FIELD_NUMBER: _ClassVar[int]
    ROLES_GRANTED_FIELD_NUMBER: _ClassVar[int]
    resource_id: str
    resource_type: _resource__client_pb2.ResourceType
    roles: _containers.RepeatedScalarFieldContainer[str]
    roles_granted: _containers.RepeatedCompositeFieldContainer[RoleAssignmentSummary]
    def __init__(self, resource_type: _Optional[_Union[_resource__client_pb2.ResourceType, str]] = ..., resource_id: _Optional[str] = ..., roles: _Optional[_Iterable[str]] = ..., roles_granted: _Optional[_Iterable[_Union[RoleAssignmentSummary, _Mapping]]] = ...) -> None: ...

class ResourceAndRoleAssignmentsV2(_message.Message):
    __slots__ = ["resource_id", "resource_type", "roles_granted"]
    RESOURCE_ID_FIELD_NUMBER: _ClassVar[int]
    RESOURCE_TYPE_FIELD_NUMBER: _ClassVar[int]
    ROLES_GRANTED_FIELD_NUMBER: _ClassVar[int]
    resource_id: str
    resource_type: _resource__client_pb2.ResourceType
    roles_granted: _containers.RepeatedCompositeFieldContainer[RoleAssignmentSummary]
    def __init__(self, resource_type: _Optional[_Union[_resource__client_pb2.ResourceType, str]] = ..., resource_id: _Optional[str] = ..., roles_granted: _Optional[_Iterable[_Union[RoleAssignmentSummary, _Mapping]]] = ...) -> None: ...

class RoleAssignmentSource(_message.Message):
    __slots__ = ["assignment_type", "principal_group_id", "principal_group_name"]
    ASSIGNMENT_TYPE_FIELD_NUMBER: _ClassVar[int]
    PRINCIPAL_GROUP_ID_FIELD_NUMBER: _ClassVar[int]
    PRINCIPAL_GROUP_NAME_FIELD_NUMBER: _ClassVar[int]
    assignment_type: RoleAssignmentType
    principal_group_id: str
    principal_group_name: str
    def __init__(self, assignment_type: _Optional[_Union[RoleAssignmentType, str]] = ..., principal_group_name: _Optional[str] = ..., principal_group_id: _Optional[str] = ...) -> None: ...

class RoleAssignmentSummary(_message.Message):
    __slots__ = ["role", "role_assignment_sources"]
    ROLE_ASSIGNMENT_SOURCES_FIELD_NUMBER: _ClassVar[int]
    ROLE_FIELD_NUMBER: _ClassVar[int]
    role: str
    role_assignment_sources: _containers.RepeatedCompositeFieldContainer[RoleAssignmentSource]
    def __init__(self, role: _Optional[str] = ..., role_assignment_sources: _Optional[_Iterable[_Union[RoleAssignmentSource, _Mapping]]] = ...) -> None: ...

class RoleAssignmentType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []



from tecton_proto.args import basic_info__client_pb2 as _basic_info__client_pb2
from tecton_proto.args import diff_options__client_pb2 as _diff_options__client_pb2
from tecton_proto.common import id__client_pb2 as _id__client_pb2
from tecton_proto.common import scaling_config__client_pb2 as _scaling_config__client_pb2
from tecton_proto.common import server_group_type__client_pb2 as _server_group_type__client_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class FeatureServerGroupArgs(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...

class ServerGroupArgs(_message.Message):
    __slots__ = ["autoscaling_config", "feature_server_group_args", "info", "options", "prevent_destroy", "provisioned_scaling_config", "server_group_id", "server_group_type", "transform_server_group_args"]
    class OptionsEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    AUTOSCALING_CONFIG_FIELD_NUMBER: _ClassVar[int]
    FEATURE_SERVER_GROUP_ARGS_FIELD_NUMBER: _ClassVar[int]
    INFO_FIELD_NUMBER: _ClassVar[int]
    OPTIONS_FIELD_NUMBER: _ClassVar[int]
    PREVENT_DESTROY_FIELD_NUMBER: _ClassVar[int]
    PROVISIONED_SCALING_CONFIG_FIELD_NUMBER: _ClassVar[int]
    SERVER_GROUP_ID_FIELD_NUMBER: _ClassVar[int]
    SERVER_GROUP_TYPE_FIELD_NUMBER: _ClassVar[int]
    TRANSFORM_SERVER_GROUP_ARGS_FIELD_NUMBER: _ClassVar[int]
    autoscaling_config: _scaling_config__client_pb2.AutoscalingConfig
    feature_server_group_args: FeatureServerGroupArgs
    info: _basic_info__client_pb2.BasicInfo
    options: _containers.ScalarMap[str, str]
    prevent_destroy: bool
    provisioned_scaling_config: _scaling_config__client_pb2.ProvisionedScalingConfig
    server_group_id: _id__client_pb2.Id
    server_group_type: _server_group_type__client_pb2.ServerGroupType
    transform_server_group_args: TransformServerGroupArgs
    def __init__(self, server_group_id: _Optional[_Union[_id__client_pb2.Id, _Mapping]] = ..., info: _Optional[_Union[_basic_info__client_pb2.BasicInfo, _Mapping]] = ..., server_group_type: _Optional[_Union[_server_group_type__client_pb2.ServerGroupType, str]] = ..., transform_server_group_args: _Optional[_Union[TransformServerGroupArgs, _Mapping]] = ..., feature_server_group_args: _Optional[_Union[FeatureServerGroupArgs, _Mapping]] = ..., prevent_destroy: bool = ..., options: _Optional[_Mapping[str, str]] = ..., autoscaling_config: _Optional[_Union[_scaling_config__client_pb2.AutoscalingConfig, _Mapping]] = ..., provisioned_scaling_config: _Optional[_Union[_scaling_config__client_pb2.ProvisionedScalingConfig, _Mapping]] = ...) -> None: ...

class ServerGroupReference(_message.Message):
    __slots__ = ["name", "server_group_id"]
    NAME_FIELD_NUMBER: _ClassVar[int]
    SERVER_GROUP_ID_FIELD_NUMBER: _ClassVar[int]
    name: str
    server_group_id: _id__client_pb2.Id
    def __init__(self, server_group_id: _Optional[_Union[_id__client_pb2.Id, _Mapping]] = ..., name: _Optional[str] = ...) -> None: ...

class TransformServerGroupArgs(_message.Message):
    __slots__ = ["environment"]
    ENVIRONMENT_FIELD_NUMBER: _ClassVar[int]
    environment: str
    def __init__(self, environment: _Optional[str] = ...) -> None: ...



import typing
from dataclasses import dataclass
from datetime import datetime
from datetime import timedelta
from typing import Any
from typing import Dict
from typing import Optional
from typing import Union

from typeguard import typechecked

import tecton_core.tecton_pendulum as pendulum
from tecton_core.errors import TectonValidationError


if typing.TYPE_CHECKING:
    import pyspark


class MaterializationContext:
    """
    MaterializationContext is a class that holds the context with relevant metadata, secrets, and resources for a materialization job.
    """

    _start_time: Optional[datetime]
    _end_time: Optional[datetime]
    _batch_schedule: timedelta
    _resources: Optional[Dict[str, Any]] = None
    _secrets: Optional[Dict[str, str]] = None

    def __init__(
        self,
        batch_schedule: Optional[timedelta] = None,
        secrets: Optional[Dict[str, str]] = None,
        resources: Optional[Dict[str, Any]] = None,
        _start_time: Optional[datetime] = None,
        _end_time: Optional[datetime] = None,
    ) -> None:
        """
        Initialize the MaterializationContext object.

        :param _start_time: The start time of the materialization job.
        :param _end_time: The end time of the materialization job.
        :param batch_schedule: The batch schedule of the Feature View.
        :param secrets: A map that maps the secret name to the resolved secret value.
        :param resources: A map that maps the resource name to the resource returned by the resource provider.
        """
        self._start_time = _start_time
        self._end_time = _end_time
        self._batch_schedule = batch_schedule
        self._secrets = secrets
        self._resources = resources

    @property
    def start_time(self) -> Optional[datetime]:
        return self._start_time

    @property
    def end_time(self) -> Optional[datetime]:
        return self._end_time

    @property
    def end_time_inclusive(self) -> Optional[datetime]:
        if self._end_time is None:
            return None

        return self._end_time - timedelta(microseconds=1)

    @property
    def batch_schedule(self) -> timedelta:
        return self._batch_schedule

    @property
    def secrets(self) -> Optional[Dict[str, str]]:
        return self._secrets

    @property
    def resources(self) -> Optional[Dict[str, Any]]:
        return self._resources

    @typechecked
    def time_filter_sql(self, timestamp_expr: str) -> str:
        """
        Returns a SQL string that filters the timestamp_expr based on the context's start_time and end_time.

        :param timestamp_expr: The timestamp expression to filter.
        """
        # Use atom string to include the timezone.
        return f"('{self.start_time.isoformat()}' <= ({timestamp_expr}) AND ({timestamp_expr}) < '{self.end_time.isoformat()}')"

    def time_filter_pyspark(self, timestamp_expr: Union[str, "pyspark.sql.Column"]) -> "pyspark.sql.Column":  # type: ignore
        """
        Returns a PySpark Column that filters the timestamp_expr based on the context's start_time and end_time.

        :param timestamp_expr: The timestamp expression to filter.
        """

        from pyspark.sql.functions import expr
        from pyspark.sql.functions import lit

        if isinstance(timestamp_expr, str):
            timestamp_expr = expr(timestamp_expr)

        # Use atom string to include the timezone.
        return (lit(self.start_time.isoformat()) <= timestamp_expr) & (timestamp_expr < lit(self.end_time.isoformat()))

    def feature_time_filter_pyspark(self, timestamp_expr: Union[str, "pyspark.sql.Column"]) -> "pyspark.sql.Column":  # type: ignore
        """
        Returns a PySpark Column that filters the timestamp_expr based on the context's start_time and end_time.

        :param timestamp_expr: The timestamp expression to filter.
        """
        return self.time_filter_pyspark(timestamp_expr)

    # Everything below is deprecated but kept for backwards compatibility since Snowflake Materialization Runtime
    # is not versioned.
    @property
    def feature_start_time(self) -> datetime:
        return self._start_time

    @property
    def feature_end_time(self) -> datetime:
        return self._end_time

    @property
    def feature_start_time_string(self) -> str:
        return self.feature_start_time.isoformat()

    @property
    def feature_end_time_string(self) -> str:
        return self.feature_end_time.isoformat()

    @typechecked
    def feature_time_filter_sql(self, timestamp_expr: str) -> str:
        """
        Returns a SQL string that filters the timestamp_expr based on the context's start_time and end_time.

        :param timestamp_expr: The timestamp expression to filter.
        """
        return self.time_filter_sql(timestamp_expr)

    @classmethod
    def _create_internal(
        cls,
        start_time: Optional[pendulum.DateTime],
        end_time: Optional[pendulum.DateTime],
        batch_schedule: Optional[pendulum.Duration],
    ) -> "MaterializationContext":
        """
        Deprecated. Should be removed with the removal of Tecton on Snowflake.
        """
        start_dt = datetime.fromtimestamp(start_time.timestamp(), pendulum.tz.UTC)
        end_dt = datetime.fromtimestamp(end_time.timestamp(), pendulum.tz.UTC)

        # should only be used in pipeline_helper
        return MaterializationContext(
            _start_time=start_dt,
            _end_time=end_dt,
            batch_schedule=batch_schedule.as_timedelta(),
        )

    @classmethod
    def _create_from_period(
        cls, feature_time_limits: Optional[pendulum.Period], batch_schedule: pendulum.Duration
    ) -> "MaterializationContext":
        feature_start_time = (
            feature_time_limits.start
            if feature_time_limits is not None
            else pendulum.from_timestamp(0, pendulum.tz.UTC)
        )
        feature_end_time = feature_time_limits.end if feature_time_limits is not None else pendulum.datetime(2100, 1, 1)

        start_dt = datetime.fromtimestamp(feature_start_time.timestamp(), pendulum.tz.UTC)
        end_dt = datetime.fromtimestamp(feature_end_time.timestamp(), pendulum.tz.UTC)

        return MaterializationContext(
            _start_time=start_dt,
            _end_time=end_dt,
            batch_schedule=batch_schedule.as_timedelta(),
        )


@dataclass
class UnboundMaterializationContext(MaterializationContext):
    """
    Deprecated. Should be removed with the removal of `context=materialization_context()`

    This is only meant for instantiation in transformation default args. Using it directly will fail.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def batch_schedule(self):
        msg = "tecton.materialization_context() must be passed in via a kwarg default only. Instantiation in function body is not allowed."
        raise TectonValidationError(msg)

    @property
    def start_time(self):
        msg = "tecton.materialization_context() must be passed in via a kwarg default only. Instantiation in function body is not allowed."
        raise TectonValidationError(msg)

    @property
    def end_time(self):
        msg = "tecton.materialization_context() must be passed in via a kwarg default only. Instantiation in function body is not allowed."
        raise TectonValidationError(msg)

    @property
    def feature_start_time(self):
        msg = "tecton.materialization_context() must be passed in via a kwarg default only. Instantiation in function body is not allowed."
        raise TectonValidationError(msg)

    @property
    def feature_end_time(self):
        msg = "tecton.materialization_context() must be passed in via a kwarg default only. Instantiation in function body is not allowed."
        raise TectonValidationError(msg)


def materialization_context():
    """
    Deprecated. Should be removed after Tecton 1.1.

    Used as a default value for a Feature View or Transformation with a materialization context parameter.

    ``context.start_time`` and ``context.end_time`` return a `datetime.datetime` object equal to the beginning and end of the period being materialized respectively. For example for a batch feature view materializing data from May 1st, 2022, ``context.start_time = datetime(2022, 5, 1)`` and ``context.end_time = datetime(2022, 5, 2)``.

    The datetimes can be used in SQL query strings directly (the datetime object will be cast to an atom-formatted timestamp string and inlined as a constant in the SQL query).

    Example usage:

    .. code-block:: python

        from tecton import batch_feature_view, materialization_context
        from datetime import datetime, timedelta

        @batch_feature_view(
            sources=[transactions],
            entities=[user],
            mode='spark_sql',
            features=[Attribute("AMOUNT", Float64)],
            batch_schedule=timedelta(days=1),
            feature_start_time=datetime(2020, 10, 10),
        )
        def user_last_transaction_amount(transactions, context=materialization_context()):
            return f'''
                SELECT
                    USER_ID,
                    AMOUNT,
                    TIMESTAMP
                FROM
                    {transactions}
                WHERE TIMESTAMP >= TO_TIMESTAMP("{context.start_time}") -- e.g. TO_TIMESTAMP("2022-05-01T00:00:00+00:00")
                    AND TIMESTAMP < TO_TIMESTAMP("{context.end_time}") -- e.g. TO_TIMESTAMP("2022-05-02T00:00:00+00:00")
                '''
    """
    dummy_time = pendulum.datetime(1970, 1, 1)
    dummy_period = timedelta()
    return UnboundMaterializationContext(
        _start_time=dummy_time,
        _end_time=dummy_time,
        batch_schedule=dummy_period,
    )



"""creates a dictionary of MDS methods to the grpc call representation

we have multiple services running in the MDS server with their own service protos
combines the methods across the different service protos to a dictionary of method to grpc call repr obj
"""

from dataclasses import dataclass
from types import ModuleType
from typing import Callable
from typing import Dict
from typing import List

from google.protobuf.empty_pb2 import Empty


@dataclass
class GrpcCall:
    """
    grpc call data representation

    Attributes:
        method: grpc method name
        request_serializer: serialization function for the grpc request proto
        response_deserializer: deserialization function for the grpc response proto
    """

    method: str
    request_serializer: Callable
    response_deserializer: Callable


def get_method_name_to_grpc_call(grpc_service_modules: List[ModuleType]) -> Dict[str, GrpcCall]:
    """
    dictionary of mds methods to grpc call repr obj
    later used by the grpc stub and the http stub for MDS py client
    """
    method_name_to_call = {}
    for module in grpc_service_modules:
        for service_name, descriptor in module.DESCRIPTOR.services_by_name.items():
            for method in descriptor.methods:
                if method.input_type.name == "Empty":
                    request_serializer = Empty.SerializeToString
                else:
                    request_serializer = find_method_message(
                        module, method.input_type.name, grpc_service_modules
                    ).SerializeToString
                if method.output_type.name == "Empty":
                    response_deserializer = Empty.FromString
                else:
                    response_deserializer = find_method_message(
                        module, method.output_type.name, grpc_service_modules
                    ).FromString
                grpc_method = f"/{descriptor.full_name}/{method.name}"
                grpc_call = GrpcCall(grpc_method, request_serializer, response_deserializer)
                if method_name_to_call.get(method.name) is None:
                    method_name_to_call[method.name] = grpc_call
                else:
                    msg = f"Method name collision for method {method.name} in service {service_name} and {method_name_to_call.get(method.name).method}"
                    raise Exception(msg)
    return method_name_to_call


def find_method_message(
    method_service_module: ModuleType, message_name: str, grpc_service_modules: List[ModuleType]
) -> object:
    """
    This function looks for a method message (input or output) with the provided message_name.
    It first looks in the provided method_service_module. If it is not found there, it then goes through all the other modules in grpc_service_modules.
    Fails with an exception if the message cannot be found in any module.
    """
    result = getattr(method_service_module, message_name, None)
    if result:
        return result
    for module in grpc_service_modules:
        if module != method_service_module:
            result = getattr(module, message_name, None)
            if result:
                return result
    msg = f"Could not find message {message_name} in method service module {method_service_module} or other provided module"
    raise Exception(msg)



from google.protobuf import timestamp_pb2 as _timestamp_pb2
from tecton_proto.auth import principal__client_pb2 as _principal__client_pb2
from tecton_proto.common import id__client_pb2 as _id__client_pb2
from tecton_proto.common import schema__client_pb2 as _schema__client_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DELETED: ModelArtifactStatus
DESCRIPTOR: _descriptor.FileDescriptor
ERROR: ModelArtifactStatus
MODEL_TYPE_UNSPECIFIED: ModelType
PENDING_FILE: ModelArtifactStatus
PENDING_SCAN: ModelArtifactStatus
PYTHON: ModelType
PYTORCH: ModelType
READY: ModelArtifactStatus
STATUS_UNSPECIFIED: ModelArtifactStatus
TECTON_TEXT_EMBEDDING: ModelType

class ModelArtifactInfo(_message.Message):
    __slots__ = ["artifact_files", "created_at", "created_by", "created_by_principal", "description", "environments", "file_hashes", "id", "input_schema", "model_config_file_path", "model_file_path", "name", "output_schema", "status", "storage_path", "tags", "type", "updated_at"]
    class FileHashesEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    class TagsEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    ARTIFACT_FILES_FIELD_NUMBER: _ClassVar[int]
    CREATED_AT_FIELD_NUMBER: _ClassVar[int]
    CREATED_BY_FIELD_NUMBER: _ClassVar[int]
    CREATED_BY_PRINCIPAL_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    ENVIRONMENTS_FIELD_NUMBER: _ClassVar[int]
    FILE_HASHES_FIELD_NUMBER: _ClassVar[int]
    ID_FIELD_NUMBER: _ClassVar[int]
    INPUT_SCHEMA_FIELD_NUMBER: _ClassVar[int]
    MODEL_CONFIG_FILE_PATH_FIELD_NUMBER: _ClassVar[int]
    MODEL_FILE_PATH_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    OUTPUT_SCHEMA_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    STORAGE_PATH_FIELD_NUMBER: _ClassVar[int]
    TAGS_FIELD_NUMBER: _ClassVar[int]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    UPDATED_AT_FIELD_NUMBER: _ClassVar[int]
    artifact_files: _containers.RepeatedScalarFieldContainer[str]
    created_at: _timestamp_pb2.Timestamp
    created_by: _principal__client_pb2.Principal
    created_by_principal: _principal__client_pb2.PrincipalBasic
    description: str
    environments: _containers.RepeatedScalarFieldContainer[str]
    file_hashes: _containers.ScalarMap[str, str]
    id: _id__client_pb2.Id
    input_schema: _schema__client_pb2.Schema
    model_config_file_path: str
    model_file_path: str
    name: str
    output_schema: _schema__client_pb2.Schema
    status: ModelArtifactStatus
    storage_path: str
    tags: _containers.ScalarMap[str, str]
    type: ModelType
    updated_at: _timestamp_pb2.Timestamp
    def __init__(self, id: _Optional[_Union[_id__client_pb2.Id, _Mapping]] = ..., name: _Optional[str] = ..., type: _Optional[_Union[ModelType, str]] = ..., description: _Optional[str] = ..., file_hashes: _Optional[_Mapping[str, str]] = ..., input_schema: _Optional[_Union[_schema__client_pb2.Schema, _Mapping]] = ..., output_schema: _Optional[_Union[_schema__client_pb2.Schema, _Mapping]] = ..., tags: _Optional[_Mapping[str, str]] = ..., storage_path: _Optional[str] = ..., status: _Optional[_Union[ModelArtifactStatus, str]] = ..., created_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., updated_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., model_file_path: _Optional[str] = ..., model_config_file_path: _Optional[str] = ..., artifact_files: _Optional[_Iterable[str]] = ..., environments: _Optional[_Iterable[str]] = ..., created_by: _Optional[_Union[_principal__client_pb2.Principal, _Mapping]] = ..., created_by_principal: _Optional[_Union[_principal__client_pb2.PrincipalBasic, _Mapping]] = ...) -> None: ...

class ModelInfo(_message.Message):
    __slots__ = ["id", "metadata_public_uri", "model_public_uri", "name"]
    ID_FIELD_NUMBER: _ClassVar[int]
    METADATA_PUBLIC_URI_FIELD_NUMBER: _ClassVar[int]
    MODEL_PUBLIC_URI_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    id: _id__client_pb2.Id
    metadata_public_uri: str
    model_public_uri: str
    name: str
    def __init__(self, id: _Optional[_Union[_id__client_pb2.Id, _Mapping]] = ..., name: _Optional[str] = ..., model_public_uri: _Optional[str] = ..., metadata_public_uri: _Optional[str] = ...) -> None: ...

class ModelArtifactStatus(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class ModelType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []



from typing import Mapping
from typing import Optional


class ResourceProviderContext:
    """
    ResourceProviderContext is a class that is used to pass context metadata
    to the `context` parameter of a Resource Provider
    """

    _secrets: Optional[Mapping[str, str]] = None

    def __init__(
        self,
        secrets: Optional[Mapping[str, str]] = None,
    ) -> None:
        """
        Initialize the ResourceProviderContext object.

        :param secrets: A map that maps the secret name to the resolved secret value.
        """
        self._secrets = secrets

    @property
    def secrets(self) -> Optional[Mapping[str, str]]:
        return self._secrets



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
An interactive shell.

This file is designed to be launched as a PYTHONSTARTUP script.
"""

import atexit
import os
import platform
import warnings

from pyspark.context import SparkContext
from pyspark.sql import SparkSession

if os.environ.get("SPARK_EXECUTOR_URI"):
    SparkContext.setSystemProperty("spark.executor.uri", os.environ["SPARK_EXECUTOR_URI"])

SparkContext._ensure_initialized()  # type: ignore

try:
    spark = SparkSession._create_shell_session()  # type: ignore
except Exception:
    import sys
    import traceback
    warnings.warn("Failed to initialize Spark session.")
    traceback.print_exc(file=sys.stderr)
    sys.exit(1)

sc = spark.sparkContext
sql = spark.sql
atexit.register(lambda: sc.stop())

# for compatibility
sqlContext = spark._wrapped
sqlCtx = sqlContext

print(r"""Welcome to
      ____              __
     / __/__  ___ _____/ /__
    _\ \/ _ \/ _ `/ __/  '_/
   /__ / .__/\_,_/_/ /_/\_\   version %s
      /_/
""" % sc.version)
print("Using Python version %s (%s, %s)" % (
    platform.python_version(),
    platform.python_build()[0],
    platform.python_build()[1]))
print("Spark context Web UI available at %s" % (sc.uiWebUrl))
print("Spark context available as 'sc' (master = %s, app id = %s)." % (sc.master, sc.applicationId))
print("SparkSession available as 'spark'.")

# The ./bin/pyspark script stores the old PYTHONSTARTUP value in OLD_PYTHONSTARTUP,
# which allows us to execute the user's PYTHONSTARTUP file:
_pythonstartup = os.environ.get('OLD_PYTHONSTARTUP')
if _pythonstartup and os.path.isfile(_pythonstartup):
    with open(_pythonstartup) as f:
        code = compile(f.read(), _pythonstartup, 'exec')
        exec(code)



import io
import time
from abc import abstractmethod
from dataclasses import dataclass
from datetime import datetime
from datetime import timedelta
from typing import TYPE_CHECKING
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import pandas

from tecton._internals import metadata_service
from tecton._internals import utils
from tecton._internals.display import Displayable
from tecton._internals.ingest_utils import upload
from tecton_core import specs
from tecton_core.arrow import PARQUET_WRITE_OPTIONS_KWARGS
from tecton_core.compute_mode import BatchComputeMode
from tecton_core.data_processing_utils import infer_pandas_timestamp
from tecton_core.errors import TectonValidationError
from tecton_core.id_helper import IdHelper
from tecton_core.query.retrieval_params import GetFeaturesForEventsParams
from tecton_core.query.retrieval_params import GetFeaturesInRangeParams
from tecton_core.schema import Schema
from tecton_proto.args.feature_view__client_pb2 import ClusterConfig
from tecton_proto.common import id__client_pb2 as id_pb2
from tecton_proto.data.materialization_status__client_pb2 import MaterializationStatus
from tecton_proto.materialization.spark_cluster__client_pb2 import TaskType
from tecton_proto.materializationjobservice.materialization_job_service__client_pb2 import CancelDatasetJobRequest
from tecton_proto.materializationjobservice.materialization_job_service__client_pb2 import CancelJobRequest
from tecton_proto.materializationjobservice.materialization_job_service__client_pb2 import (
    CompleteDataframeUploadRequest,
)
from tecton_proto.materializationjobservice.materialization_job_service__client_pb2 import GetDataframeUploadUrlRequest
from tecton_proto.materializationjobservice.materialization_job_service__client_pb2 import GetDatasetJobRequest
from tecton_proto.materializationjobservice.materialization_job_service__client_pb2 import GetJobRequest
from tecton_proto.materializationjobservice.materialization_job_service__client_pb2 import GetLatestReadyTimeRequest
from tecton_proto.materializationjobservice.materialization_job_service__client_pb2 import GetLatestReadyTimeResponse
from tecton_proto.materializationjobservice.materialization_job_service__client_pb2 import JobAttempt
from tecton_proto.materializationjobservice.materialization_job_service__client_pb2 import ListJobsRequest
from tecton_proto.materializationjobservice.materialization_job_service__client_pb2 import (
    MaterializationJob as MaterializationJobProto,
)
from tecton_proto.materializationjobservice.materialization_job_service__client_pb2 import MaterializationJobRequest
from tecton_proto.materializationjobservice.materialization_job_service__client_pb2 import StartDatasetJobRequest
from tecton_proto.metadataservice.metadata_service__client_pb2 import GetJobStatusRequest
from tecton_proto.metadataservice.metadata_service__client_pb2 import GetMaterializationStatusRequest
from tecton_proto.metadataservice.metadata_service__client_pb2 import GetSavedFeatureDataFrameRequest


if TYPE_CHECKING:
    from tecton.framework.dataset import SavedDataset


class JobTimeoutException(Exception):
    pass


class JobFailedException(Exception):
    pass


WAIT_INTERVAL = timedelta(seconds=60)


@dataclass
class MaterializationAttempt:
    """
    Data representation of the materialization job attempt.

    Materialization job may have multiple attempts to materialize features.

    :param id: ID string of the materialization attempt.
    :param run_url: URL to track materialization attempt.
    :param state: State of the materialization attempt.
    :param created_at: Materialization attempt creation timestamp.
    :param updated_at: Materialization attempt update timestamp.
    """

    id: str
    run_url: str
    state: str
    created_at: datetime
    updated_at: datetime

    @classmethod
    def _from_proto(cls, proto: JobAttempt):
        created_at = datetime.utcfromtimestamp(proto.created_at.seconds)
        updated_at = datetime.utcfromtimestamp(proto.updated_at.seconds)
        return cls(id=proto.id, run_url=proto.run_url, state=proto.state, created_at=created_at, updated_at=updated_at)


@dataclass
class TectonJob:
    """
    Base class for Tecton jobs executed remotely by Tecton compute
    """

    id: str
    workspace: str
    feature_view: Optional[str]
    feature_service: Optional[str]
    state: str
    created_at: datetime
    updated_at: datetime
    attempts: List[MaterializationAttempt]
    next_attempt_at: Optional[datetime]
    job_type: str

    @classmethod
    def _from_proto(cls, proto: MaterializationJobProto):
        if proto.HasField("saved_feature_data_frame"):
            return DatasetJob._from_proto(proto)

        return MaterializationJob._from_proto(proto)

    def cancel(self) -> "TectonJob":
        """
        Cancels the scheduled or running job by the job identifier.
        Once cancelled, a job will not be retried further.

        Job run state will be set to MANUAL_CANCELLATION_REQUESTED.
        Note that cancellation is asynchronous, so it may take some time for the cancellation to complete.
        If job run is already in MANUAL_CANCELLATION_REQUESTED or in a terminal state then it'll return the job.

        :param job_id: ID string of the materialization job.
        :return: `JobData` object for the cancelled job.
        """
        return cancel_job(self.workspace, self.id, feature_view=self.feature_view, feature_service=self.feature_service)

    @abstractmethod
    def get_status_for_display(self) -> Displayable: ...


@dataclass
class MaterializationJob(TectonJob):
    """
    Representation of the materialization job

    :param id: ID string of the materialization job.
    :param workspace: Name of the project workspace.
    :param feature_view: Name of the Feature View (if job attached to a feature view).
    :param feature_service: Name of the Feature Service (if job attached to a feature service).
    :param state: State of the materialization job.
    :param online: Whether the job materializes features to the online store.
    :param offline: Whether the job materializes features to the offline store.
    :param start_time: Start timestamp of the batch materialization window.
    :param end_time: End timestamp of the batch materialization window.
    :param created_at: Job creation timestamp.
    :param updated_at: Job update timestamp.
    :param attempts: Materialization attempts. List of `MaterializationAttemptData`
    :param next_attempt_at: If job needs another attempt, Start timestamp the next materialization attempt.
    :param job_type: Type of materialization. One of 'BATCH' or 'STREAM'.
    """

    online: bool
    offline: bool
    start_time: Optional[datetime]
    end_time: Optional[datetime]

    @classmethod
    def _from_proto(cls, proto: MaterializationJobProto):
        start_time = datetime.utcfromtimestamp(proto.start_time.seconds) if proto.HasField("start_time") else None
        end_time = datetime.utcfromtimestamp(proto.end_time.seconds) if proto.HasField("end_time") else None
        created_at = datetime.utcfromtimestamp(proto.created_at.seconds)
        updated_at = datetime.utcfromtimestamp(proto.updated_at.seconds)
        attempts = [MaterializationAttempt._from_proto(attempt_proto) for attempt_proto in proto.attempts]
        next_attempt_at = (
            datetime.utcfromtimestamp(proto.next_attempt_at.seconds) if proto.HasField("next_attempt_at") else None
        )
        return cls(
            id=proto.id,
            workspace=proto.workspace,
            feature_view=proto.feature_view if proto.HasField("feature_view") else None,
            feature_service=proto.feature_service if proto.HasField("feature_service") else None,
            state=proto.state,
            online=proto.online,
            offline=proto.offline,
            start_time=start_time,
            end_time=end_time,
            created_at=created_at,
            updated_at=updated_at,
            attempts=attempts,
            next_attempt_at=next_attempt_at,
            job_type=proto.job_type,
        )

    def get_status_for_display(self) -> Displayable:
        """
        Returns a user-friendly representation of this job's status (with attempts).
        """
        status = get_job_status(IdHelper.from_string(self.id), self.workspace)
        column_names, materialization_status_rows = utils.format_materialization_attempts(
            status.materialization_attempts,
            verbose=True,
        )
        return _create_materialization_table(column_names, materialization_status_rows)


@dataclass
class DatasetJob(TectonJob):
    """
    Representation of dataset generation job

    :param id: ID string of the materialization job.
    :param workspace: Name of the project workspace.
    :param feature_view: Name of the Feature View (if job attached to a feature view).
    :param feature_service: Name of the Feature Service (if job attached to a feature service).
    :param state: State of the materialization job.
    :param created_at: Job creation timestamp.
    :param updated_at: Job update timestamp.
    :param attempts: Materialization attempts. List of `MaterializationAttemptData`
    :param next_attempt_at: If job needs another attempt, Start timestamp the next materialization attempt.
    :param job_type: Type of materialization. For now only one value is expected 'DATASET_GENERATION'.
    :param dataset: Name of the Dataset.
    """

    dataset: str

    @classmethod
    def _from_proto(cls, proto: MaterializationJobProto):
        created_at = datetime.utcfromtimestamp(proto.created_at.seconds)
        updated_at = datetime.utcfromtimestamp(proto.updated_at.seconds)
        attempts = [MaterializationAttempt._from_proto(attempt_proto) for attempt_proto in proto.attempts]
        next_attempt_at = (
            datetime.utcfromtimestamp(proto.next_attempt_at.seconds) if proto.HasField("next_attempt_at") else None
        )
        return cls(
            id=proto.id,
            workspace=proto.workspace,
            feature_view=proto.feature_view if proto.HasField("feature_view") else None,
            feature_service=proto.feature_service if proto.HasField("feature_service") else None,
            dataset=proto.saved_feature_data_frame,
            state=proto.state,
            created_at=created_at,
            updated_at=updated_at,
            attempts=attempts,
            next_attempt_at=next_attempt_at,
            job_type=proto.job_type,
        )

    def wait_for_completion(self, timeout: Optional[timedelta] = None):
        """
        Blocks until the job has been completed.

        :param timeout: (Optional) timeout for this function.
            An exception is raised if the job does not complete within the specified time.
        """
        self.state = wait_for_dataset_job(self.workspace, self.id, self.dataset, timeout=timeout).state

    def get_dataset(self) -> Optional["SavedDataset"]:
        """
        Retrieves the dataset instance after the job is completed.

        :return: `SavedDataset` object if the job has finished successfully. Otherwise, returns None.
        """
        from tecton.framework.dataset import SavedDataset

        self.state = get_dataset_job(self.workspace, self.id, self.dataset).state

        if self.state != "SUCCESS":
            return

        request = GetSavedFeatureDataFrameRequest(saved_feature_dataframe_name=self.dataset, workspace=self.workspace)
        response = metadata_service.instance().GetSavedFeatureDataFrame(request)
        return SavedDataset._from_proto(response.saved_feature_dataframe)

    def get_status_for_display(self) -> Displayable:
        """
        Returns a user-friendly representation of this job's status (with attempts).
        """
        status = get_job_status(IdHelper.from_string(self.id), self.workspace)
        column_names, materialization_status_rows = utils.format_materialization_attempts(
            status.materialization_attempts,
            override_columns=[
                "TYPE",
                "STATUS",
                "ATTEMPT",
                "ENV_VERSION",
                "TERMINATION_REASON",
                "STATE_MESSAGE",
                "TASK_SCHEDULED_AT",
                "JOB_LOGS",
            ],
        )
        return _create_materialization_table(column_names, materialization_status_rows)


def trigger_materialization_job(
    feature_view: str,
    workspace: str,
    start_time: datetime,
    end_time: datetime,
    online: bool,
    offline: bool,
    use_tecton_managed_retries: bool = True,
    overwrite: bool = False,
) -> str:
    """
    Starts a batch materialization job for this Feature View.

    :param start_time: The job will materialize feature values between the start_time and end_time.
    :param end_time: The job will materialize feature values between the start_time and end_time.
    :param online: Whether the job will materialize features to the online store.
        The Feature View must be configured with online=True in order to materialize features online.
    :param offline: Whether the job will materialize features to the offline store.
        The Feature View must be configured with offline=True in order to materialize features offline.
    :param use_tecton_managed_retries: If enabled, Tecton will automatically retry failed attempts.
        Disable to manage your own retry behavior.
    :param overwrite: If enabled, you will be able to run materialization jobs for periods that previously have materialized data.
        This operation can be sensitive for feature views with existing materialized online data.
        For the offline store, all previously materialized data between the start time and end time will be dropped.
        For the online store, all previous data will remain, but may be overwritten by this job.
    :return: ID string of the created materialization job.
    :raises TectonValidationError: If job params are not valid.
    """
    request = MaterializationJobRequest()
    request.feature_view = feature_view
    request.workspace = workspace
    request.start_time.FromDatetime(start_time)
    request.end_time.FromDatetime(end_time)
    request.online = online
    request.offline = offline
    request.use_tecton_managed_retries = use_tecton_managed_retries
    request.overwrite = overwrite

    mds_instance = metadata_service.instance()
    response = mds_instance.SubmitMaterializationJob(request)
    return response.job.id


def list_jobs(
    workspace: str, feature_view: Optional[str] = None, feature_service: Optional[str] = None
) -> List[TectonJob]:
    """
    Retrieves the list of all jobs for the provided feature view or feature service.
    :return: List of `JobData` objects.
    """
    assert (feature_view is not None) ^ (
        feature_service is not None
    ), "Either feature_view or feature_service must be provided. Not both."

    request = ListJobsRequest()
    if feature_view:
        request.feature_view = feature_view
    if feature_service:
        request.feature_service = feature_service
    request.workspace = workspace

    mds_instance = metadata_service.instance()
    response = mds_instance.ListMaterializationJobs(request)
    return [TectonJob._from_proto(job) for job in response.jobs]


def get_job(
    workspace: str, job_id: str, feature_view: Optional[str] = None, feature_service: Optional[str] = None
) -> TectonJob:
    """
    Retrieves data about the specified job for the provided feature view or feature service.

    This data includes information about job attempts.

    :param job_id: ID string of the materialization job.
    :return: `JobData` object for the job.
    """
    assert (feature_view is not None) ^ (
        feature_service is not None
    ), "Either feature_view or feature_service must be provided. Not both."

    request = GetJobRequest()
    if feature_view:
        request.feature_view = feature_view
    if feature_service:
        request.feature_service = feature_service
    request.workspace = workspace
    request.job_id = job_id

    mds_instance = metadata_service.instance()
    response = mds_instance.GetMaterializationJob(request)
    return TectonJob._from_proto(response.job)


def get_latest_ready_time(feature_view: str, workspace: str) -> GetLatestReadyTimeResponse:
    request = GetLatestReadyTimeRequest()
    request.feature_view = feature_view
    request.workspace = workspace

    mds_instance = metadata_service.instance()
    return mds_instance.GetLatestReadyTime(request)


def cancel_job(
    workspace: str, job_id: str, feature_view: Optional[str] = None, feature_service: Optional[str] = None
) -> TectonJob:
    """
    Cancels the scheduled or running job by the job identifier.
    Once cancelled, a job will not be retried further.

    Job run state will be set to `MANUAL_CANCELLATION_REQUESTED`.
    Note that cancellation is asynchronous, so it may take some time for the cancellation to complete.
    If job run is already in `MANUAL_CANCELLATION_REQUESTED` or in a terminal state then it'll return the job.

    :param job_id: ID string of the materialization job.
    :return: `JobData` object for the cancelled job.
    """
    assert (feature_view is not None) ^ (
        feature_service is not None
    ), "Either feature_view or feature_service must be provided. Not both."

    request = CancelJobRequest()
    if feature_view:
        request.feature_view = feature_view
    if feature_service:
        request.feature_service = feature_service
    request.workspace = workspace
    request.job_id = job_id

    mds_instance = metadata_service.instance()
    response = mds_instance.CancelMaterializationJob(request)
    return TectonJob._from_proto(response.job)


def wait_for_materialization_job(
    workspace: str,
    job_id: str,
    feature_view: Optional[str] = None,
    feature_service: Optional[str] = None,
    timeout: Optional[timedelta] = None,
) -> TectonJob:
    """
    Blocks until the specified job has been completed.

    :param job_id: ID string of the materialization job.
    :param timeout: (Optional) timeout for this function.
        An exception is raised if the job does not complete within the specified time.
    :return: `JobData` object for the successful job.
    :raises JobTimeoutException:
        If timeout param is specified and job does not complete within the specified time.
    :raises MaterializationJobFailedException: If materialization job did not reach a successful state.
    """
    return _wait_for_job(
        lambda: get_job(workspace, job_id, feature_view, feature_service),
        job_id,
        timeout,
    )


def _wait_for_job(
    job_retriever: Callable[[], TectonJob],
    job_id: str,
    timeout: timedelta,
    max_retries: int = 10,
) -> TectonJob:
    wait_start_time = datetime.now()
    retries = 0
    while True:
        try:
            job_data = job_retriever()
        except IOError:
            retries += 1
            if retries > max_retries:
                raise

            continue
        else:
            # reset retries on successful retrieval
            retries = 0

        run_state = job_data.state

        if run_state == "SUCCESS":
            return job_data
        elif timeout and ((datetime.now() - wait_start_time) > timeout):
            msg = f"job {job_id} timed out, last job state {run_state}"
            raise JobTimeoutException(msg)
        elif run_state == "RUNNING":
            time.sleep(WAIT_INTERVAL.total_seconds())
        else:
            msg = f"job {job_id} failed, last job state {run_state}"
            raise JobFailedException(msg)


def get_materialization_status_response(id_proto: id_pb2.Id, workspace: str) -> MaterializationStatus:
    """Returns MaterializationStatus proto for the FeatureView."""
    request = GetMaterializationStatusRequest()
    request.feature_package_id.CopyFrom(id_proto)
    request.workspace = workspace

    response = metadata_service.instance().GetMaterializationStatus(request)
    return response.materialization_status


def get_job_status(task_id: id_pb2.Id, workspace: str) -> MaterializationStatus:
    request = GetJobStatusRequest()
    request.task_id.CopyFrom(task_id)
    request.workspace = workspace

    response = metadata_service.instance().GetJobStatus(request)
    return response.status


@dataclass
class SpineInput:
    path: str
    timestamp_key: str
    column_names: List[str]


@dataclass
class DateTimeRangeInput:
    start_time: datetime
    end_time: datetime
    max_loopback: Optional[timedelta] = None
    entities_path: Optional[str] = None


def _build_start_dataset_job_request(
    fco: Union[specs.FeatureServiceSpec, specs.FeatureViewSpec],
    input_: Union[SpineInput, DateTimeRangeInput],
    dataset: str,
    compute_mode: BatchComputeMode,
    output_schema: Schema,
    from_source: Optional[bool] = None,
    cluster_config: Optional[ClusterConfig] = None,
    tecton_materialization_runtime: Optional[str] = None,
    environment: Optional[str] = None,
    extra_config: Optional[Dict[str, Any]] = None,
) -> StartDatasetJobRequest:
    """This function shared between `start_dataset_job` and local integration tests"""
    request = StartDatasetJobRequest()
    request.workspace = fco.workspace
    request.compute_mode = compute_mode.value
    request.dataset_name = dataset
    request.expected_schema.CopyFrom(output_schema.to_proto())
    if from_source is not None:
        request.from_source = from_source

    if isinstance(fco, specs.FeatureViewSpec):
        request.feature_view_id.CopyFrom(fco.id_proto)
    elif isinstance(fco, specs.FeatureServiceSpec):
        request.feature_service_id.CopyFrom(fco.id_proto)
    else:
        msg = f"unexpected fco type {type(fco)} for start_dataset_job"
        raise RuntimeError(msg)

    if isinstance(input_, SpineInput):
        request.spine.path = input_.path

        request.spine.timestamp_key = input_.timestamp_key
        request.spine.column_names.extend(input_.column_names)
    elif isinstance(input_, DateTimeRangeInput):
        request.datetime_range.start.FromDatetime(input_.start_time)
        request.datetime_range.end.FromDatetime(input_.end_time)
        if input_.max_loopback:
            request.datetime_range.max_loopback.FromTimedelta(input_.max_loopback)
        if input_.entities_path:
            request.datetime_range.entities_path = input_.entities_path
    else:
        msg = f"unexpected input type {type(input_)} for start_dataset_jon"
        raise RuntimeError(msg)

    if cluster_config:
        request.cluster_config.CopyFrom(cluster_config)

    if tecton_materialization_runtime:
        request.tecton_runtime = tecton_materialization_runtime

    if environment:
        request.environment = environment

    if extra_config:
        request.extra_config.update({k: str(v) for k, v in extra_config.items()})
    return request


def start_dataset_job(
    fco: Union[specs.FeatureServiceSpec, specs.FeatureViewSpec],
    input_: Union[SpineInput, DateTimeRangeInput],
    dataset: str,
    compute_mode: BatchComputeMode,
    output_schema: Schema,
    from_source: Optional[bool] = None,
    cluster_config: Optional[ClusterConfig] = None,
    tecton_materialization_runtime: Optional[str] = None,
    environment: Optional[str] = None,
    extra_config: Optional[Dict[str, Any]] = None,
) -> DatasetJob:
    request = _build_start_dataset_job_request(
        fco,
        input_,
        dataset,
        compute_mode,
        output_schema,
        from_source,
        cluster_config,
        tecton_materialization_runtime,
        environment,
        extra_config,
    )

    response = metadata_service.instance().StartDatasetJob(request)
    return DatasetJob._from_proto(response.job)


def get_dataset_job(workspace: str, job_id: str, dataset: str):
    request = GetDatasetJobRequest()
    request.workspace = workspace
    request.saved_feature_data_frame = dataset
    request.job_id = job_id

    response = metadata_service.instance().GetDatasetJob(request)
    return DatasetJob._from_proto(response.job)


def wait_for_dataset_job(workspace: str, job_id: str, dataset: str, timeout: Optional[timedelta] = None):
    return _wait_for_job(
        lambda: get_dataset_job(workspace, job_id, dataset),
        job_id,
        timeout,
    )


def cancel_dataset_job(workspace: str, job_id: str, dataset: str) -> DatasetJob:
    request = CancelDatasetJobRequest()
    request.workspace = workspace
    request.saved_feature_data_frame = dataset
    request.job_id = job_id

    response = metadata_service.instance().CancelDatasetJob(request)
    return DatasetJob._from_proto(response.job)


def upload_dataframe_for_dataset_job(
    df: Union["pandas.DataFrame", str], fco: Union[specs.FeatureServiceSpec, specs.FeatureViewSpec], param_name: str
) -> str:
    if isinstance(df, str):
        if not df.startswith("s3://"):
            msg = f"Provide a valid S3 URI for `{param_name}` parameter"
            raise TectonValidationError(msg)
        return df
    elif not isinstance(df, pandas.DataFrame):
        msg = f"`{param_name}` parameter must be a pandas.DataFrame or an S3 URI"
        raise TectonValidationError(msg)

    # Step 1 : Create the upload request
    start_upload_request = GetDataframeUploadUrlRequest()
    workspace = fco.workspace
    start_upload_request.workspace = workspace
    if isinstance(fco, specs.FeatureViewSpec):
        start_upload_request.feature_view = fco.name
    elif isinstance(fco, specs.FeatureServiceSpec):
        start_upload_request.feature_service = fco.name
    else:
        msg = f"unexpected fco type {type(fco)}"
        raise RuntimeError(msg)
    start_upload_request.task_type = TaskType.DATASET_GENERATION

    # Step 2 : Get the upload url response
    start_upload_response = metadata_service.instance().GetDataframeUploadUrl(start_upload_request)
    out_buffer = io.BytesIO()
    df.to_parquet(out_buffer, index=False, engine="pyarrow", **PARQUET_WRITE_OPTIONS_KWARGS)
    buffer_size = out_buffer.tell()
    out_buffer.seek(0)

    # Step 3: Upload (in chunks)
    upload_id = start_upload_response.upload_id
    key = start_upload_response.key
    upload_response_map = upload(
        workspace=workspace,
        key=key,
        upload_id=upload_id,
        out_buffer=out_buffer,
        buffer_size=buffer_size,
    )
    # Step 4: Complete the upload request
    complete_upload_request = CompleteDataframeUploadRequest(
        workspace=workspace,
        key=key,
        upload_id=upload_id,
        part_etags=upload_response_map,
    )
    complete_upload_response = metadata_service.instance().CompleteDataframeUpload(complete_upload_request)
    return str(complete_upload_response.key)


def _create_materialization_table(column_names: List[str], materialization_status_rows: List[List]) -> Displayable:
    # Setting `max_width=0` creates a table with an unlimited width.
    table = Displayable.from_table(headings=column_names, rows=materialization_status_rows, max_width=0)
    # Align columns in the middle horizontally
    table._text_table.set_cols_align(["c" for _ in range(len(column_names))])

    return table


def get_materialization_status_for_display(
    id_proto: id_pb2.Id, workspace: str, verbose: bool, limit: int, sort_columns: Optional[str], errors_only: bool
) -> Displayable:
    materialization_attempts = get_materialization_status_response(id_proto, workspace).materialization_attempts
    column_names, materialization_status_rows = utils.format_materialization_attempts(
        materialization_attempts, verbose, limit, sort_columns, errors_only
    )

    return _create_materialization_table(column_names, materialization_status_rows)


def retrieval_params_to_job_input(
    params: Union[GetFeaturesInRangeParams, GetFeaturesForEventsParams],
    fco: Union[specs.FeatureServiceSpec, specs.FeatureViewSpec],
):
    if isinstance(params, GetFeaturesInRangeParams):
        input_ = DateTimeRangeInput(
            start_time=params.start_time,
            end_time=params.end_time,
            max_loopback=params.max_lookback,
        )
        if params.entities is not None:
            input_.entities_path = upload_dataframe_for_dataset_job(params.entities, fco, param_name="entities")

        return input_
    if isinstance(params, GetFeaturesForEventsParams):
        return SpineInput(
            path=upload_dataframe_for_dataset_job(params.events, fco, param_name="events"),
            column_names=params.events.columns if isinstance(params.events, pandas.DataFrame) else [],
            timestamp_key=params.timestamp_key or infer_pandas_timestamp(params.events),
        )

    msg = f"Unsupported request context type: {type(params)}"
    raise RuntimeError(msg)



from tecton_proto.common import data_type__client_pb2 as _data_type__client_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

ADDITION: OperationType
AND: OperationType
CENTURY: DatePart
COALESCE: OperationType
DATE_DIFF: OperationType
DATE_PART_UNSPECIFIED: DatePart
DAY: DatePart
DECADE: DatePart
DESCRIPTOR: _descriptor.FileDescriptor
DIVISION: OperationType
EQUALS: OperationType
GREATER_THAN: OperationType
GREATER_THAN_EQUALS: OperationType
HOUR: DatePart
LESS_THAN: OperationType
LESS_THAN_EQUALS: OperationType
MICROSECONDS: DatePart
MILLENNIUM: DatePart
MILLISECONDS: DatePart
MINUTE: DatePart
MONTH: DatePart
MULTIPLICATION: OperationType
NOT: OperationType
NOT_EQUALS: OperationType
OPERATION_UNSPECIFIED: OperationType
OR: OperationType
QUARTER: DatePart
SECOND: DatePart
SUBTRACTION: OperationType
WEEK: DatePart
YEAR: DatePart

class AbstractSyntaxTreeNode(_message.Message):
    __slots__ = ["case_statement", "column_reference", "date_part", "dtype", "literal_value", "operation"]
    CASE_STATEMENT_FIELD_NUMBER: _ClassVar[int]
    COLUMN_REFERENCE_FIELD_NUMBER: _ClassVar[int]
    DATE_PART_FIELD_NUMBER: _ClassVar[int]
    DTYPE_FIELD_NUMBER: _ClassVar[int]
    LITERAL_VALUE_FIELD_NUMBER: _ClassVar[int]
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    case_statement: CaseStatement
    column_reference: str
    date_part: DatePart
    dtype: _data_type__client_pb2.DataType
    literal_value: LiteralValue
    operation: Operation
    def __init__(self, dtype: _Optional[_Union[_data_type__client_pb2.DataType, _Mapping]] = ..., literal_value: _Optional[_Union[LiteralValue, _Mapping]] = ..., column_reference: _Optional[str] = ..., operation: _Optional[_Union[Operation, _Mapping]] = ..., date_part: _Optional[_Union[DatePart, str]] = ..., case_statement: _Optional[_Union[CaseStatement, _Mapping]] = ...) -> None: ...

class CaseStatement(_message.Message):
    __slots__ = ["else_clause", "when_clauses"]
    ELSE_CLAUSE_FIELD_NUMBER: _ClassVar[int]
    WHEN_CLAUSES_FIELD_NUMBER: _ClassVar[int]
    else_clause: ElseClause
    when_clauses: _containers.RepeatedCompositeFieldContainer[WhenClause]
    def __init__(self, when_clauses: _Optional[_Iterable[_Union[WhenClause, _Mapping]]] = ..., else_clause: _Optional[_Union[ElseClause, _Mapping]] = ...) -> None: ...

class ElseClause(_message.Message):
    __slots__ = ["result"]
    RESULT_FIELD_NUMBER: _ClassVar[int]
    result: AbstractSyntaxTreeNode
    def __init__(self, result: _Optional[_Union[AbstractSyntaxTreeNode, _Mapping]] = ...) -> None: ...

class LiteralValue(_message.Message):
    __slots__ = ["bool_value", "float32_value", "float64_value", "int64_value", "null_value", "string_value"]
    BOOL_VALUE_FIELD_NUMBER: _ClassVar[int]
    FLOAT32_VALUE_FIELD_NUMBER: _ClassVar[int]
    FLOAT64_VALUE_FIELD_NUMBER: _ClassVar[int]
    INT64_VALUE_FIELD_NUMBER: _ClassVar[int]
    NULL_VALUE_FIELD_NUMBER: _ClassVar[int]
    STRING_VALUE_FIELD_NUMBER: _ClassVar[int]
    bool_value: bool
    float32_value: float
    float64_value: float
    int64_value: int
    null_value: NullLiteralValue
    string_value: str
    def __init__(self, float32_value: _Optional[float] = ..., float64_value: _Optional[float] = ..., int64_value: _Optional[int] = ..., bool_value: bool = ..., string_value: _Optional[str] = ..., null_value: _Optional[_Union[NullLiteralValue, _Mapping]] = ...) -> None: ...

class NullLiteralValue(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...

class Operation(_message.Message):
    __slots__ = ["operands", "operation"]
    OPERANDS_FIELD_NUMBER: _ClassVar[int]
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operands: _containers.RepeatedCompositeFieldContainer[AbstractSyntaxTreeNode]
    operation: OperationType
    def __init__(self, operation: _Optional[_Union[OperationType, str]] = ..., operands: _Optional[_Iterable[_Union[AbstractSyntaxTreeNode, _Mapping]]] = ...) -> None: ...

class WhenClause(_message.Message):
    __slots__ = ["condition", "result"]
    CONDITION_FIELD_NUMBER: _ClassVar[int]
    RESULT_FIELD_NUMBER: _ClassVar[int]
    condition: AbstractSyntaxTreeNode
    result: AbstractSyntaxTreeNode
    def __init__(self, condition: _Optional[_Union[AbstractSyntaxTreeNode, _Mapping]] = ..., result: _Optional[_Union[AbstractSyntaxTreeNode, _Mapping]] = ...) -> None: ...

class OperationType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class DatePart(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []



from google.protobuf import descriptor_pb2 as _descriptor_pb2
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional, Union as _Union

ANALYTICS_FIELD_NUMBER: _ClassVar[int]
DESCRIPTOR: _descriptor.FileDescriptor
NOOP_FILTER: AnalyticsFilterType
PSEUDONYMIZE: AnalyticsFilterType
analytics: _descriptor.FieldDescriptor

class Analytics(_message.Message):
    __slots__ = ["filter_type"]
    FILTER_TYPE_FIELD_NUMBER: _ClassVar[int]
    filter_type: AnalyticsFilterType
    def __init__(self, filter_type: _Optional[_Union[AnalyticsFilterType, str]] = ...) -> None: ...

class AnalyticsFilterType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []



from google.protobuf import timestamp_pb2 as _timestamp_pb2
from tecton_proto.auth import principal__client_pb2 as _principal__client_pb2
from tecton_proto.common import id__client_pb2 as _id__client_pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class TectonApiKeyDto(_message.Message):
    __slots__ = ["archived", "created_at", "created_by", "creator", "description", "expires_at", "hashed_key", "id", "is_active", "is_admin", "is_system_managed", "is_temporary", "name", "obscured_key"]
    ARCHIVED_FIELD_NUMBER: _ClassVar[int]
    CREATED_AT_FIELD_NUMBER: _ClassVar[int]
    CREATED_BY_FIELD_NUMBER: _ClassVar[int]
    CREATOR_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    EXPIRES_AT_FIELD_NUMBER: _ClassVar[int]
    HASHED_KEY_FIELD_NUMBER: _ClassVar[int]
    ID_FIELD_NUMBER: _ClassVar[int]
    IS_ACTIVE_FIELD_NUMBER: _ClassVar[int]
    IS_ADMIN_FIELD_NUMBER: _ClassVar[int]
    IS_SYSTEM_MANAGED_FIELD_NUMBER: _ClassVar[int]
    IS_TEMPORARY_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    OBSCURED_KEY_FIELD_NUMBER: _ClassVar[int]
    archived: bool
    created_at: _timestamp_pb2.Timestamp
    created_by: str
    creator: _principal__client_pb2.Principal
    description: str
    expires_at: _timestamp_pb2.Timestamp
    hashed_key: str
    id: _id__client_pb2.Id
    is_active: bool
    is_admin: bool
    is_system_managed: bool
    is_temporary: bool
    name: str
    obscured_key: str
    def __init__(self, id: _Optional[_Union[_id__client_pb2.Id, _Mapping]] = ..., hashed_key: _Optional[str] = ..., description: _Optional[str] = ..., archived: bool = ..., created_by: _Optional[str] = ..., obscured_key: _Optional[str] = ..., is_admin: bool = ..., name: _Optional[str] = ..., is_active: bool = ..., creator: _Optional[_Union[_principal__client_pb2.Principal, _Mapping]] = ..., created_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., is_system_managed: bool = ..., is_temporary: bool = ..., expires_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...) -> None: ...



from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor
PRINCIPAL_TYPE_GROUP: PrincipalType
PRINCIPAL_TYPE_SERVICE_ACCOUNT: PrincipalType
PRINCIPAL_TYPE_UNSPECIFIED: PrincipalType
PRINCIPAL_TYPE_USER: PrincipalType
PRINCIPAL_TYPE_WORKSPACE: PrincipalType

class GroupBasic(_message.Message):
    __slots__ = ["id", "name"]
    ID_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    id: str
    name: str
    def __init__(self, id: _Optional[str] = ..., name: _Optional[str] = ...) -> None: ...

class Principal(_message.Message):
    __slots__ = ["id", "principal_type"]
    ID_FIELD_NUMBER: _ClassVar[int]
    PRINCIPAL_TYPE_FIELD_NUMBER: _ClassVar[int]
    id: str
    principal_type: PrincipalType
    def __init__(self, principal_type: _Optional[_Union[PrincipalType, str]] = ..., id: _Optional[str] = ...) -> None: ...

class PrincipalBasic(_message.Message):
    __slots__ = ["group", "service_account", "user", "workspace"]
    GROUP_FIELD_NUMBER: _ClassVar[int]
    SERVICE_ACCOUNT_FIELD_NUMBER: _ClassVar[int]
    USER_FIELD_NUMBER: _ClassVar[int]
    WORKSPACE_FIELD_NUMBER: _ClassVar[int]
    group: GroupBasic
    service_account: ServiceAccountBasic
    user: UserBasic
    workspace: WorkspaceBasic
    def __init__(self, user: _Optional[_Union[UserBasic, _Mapping]] = ..., service_account: _Optional[_Union[ServiceAccountBasic, _Mapping]] = ..., group: _Optional[_Union[GroupBasic, _Mapping]] = ..., workspace: _Optional[_Union[WorkspaceBasic, _Mapping]] = ...) -> None: ...

class ServiceAccountBasic(_message.Message):
    __slots__ = ["creator", "description", "id", "is_active", "name", "owner"]
    CREATOR_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    ID_FIELD_NUMBER: _ClassVar[int]
    IS_ACTIVE_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    OWNER_FIELD_NUMBER: _ClassVar[int]
    creator: Principal
    description: str
    id: str
    is_active: bool
    name: str
    owner: PrincipalBasic
    def __init__(self, id: _Optional[str] = ..., name: _Optional[str] = ..., description: _Optional[str] = ..., is_active: bool = ..., creator: _Optional[_Union[Principal, _Mapping]] = ..., owner: _Optional[_Union[PrincipalBasic, _Mapping]] = ...) -> None: ...

class UserBasic(_message.Message):
    __slots__ = ["first_name", "last_name", "login_email", "okta_id"]
    FIRST_NAME_FIELD_NUMBER: _ClassVar[int]
    LAST_NAME_FIELD_NUMBER: _ClassVar[int]
    LOGIN_EMAIL_FIELD_NUMBER: _ClassVar[int]
    OKTA_ID_FIELD_NUMBER: _ClassVar[int]
    first_name: str
    last_name: str
    login_email: str
    okta_id: str
    def __init__(self, okta_id: _Optional[str] = ..., first_name: _Optional[str] = ..., last_name: _Optional[str] = ..., login_email: _Optional[str] = ...) -> None: ...

class WorkspaceBasic(_message.Message):
    __slots__ = ["name"]
    NAME_FIELD_NUMBER: _ClassVar[int]
    name: str
    def __init__(self, name: _Optional[str] = ...) -> None: ...

class PrincipalType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []



import re
from typing import Dict
from typing import List
from typing import Union

from tecton._internals.errors import ExpressionParsingError
from tecton.types import SdkDataType
from tecton.vendor.sql_expresssion_parser import Lark_StandAlone
from tecton.vendor.sql_expresssion_parser import LexError
from tecton.vendor.sql_expresssion_parser import ParseError
from tecton.vendor.sql_expresssion_parser import Token
from tecton.vendor.sql_expresssion_parser import Transformer
from tecton.vendor.sql_expresssion_parser import VisitError
from tecton_core.data_types import BoolType
from tecton_core.data_types import Float32Type
from tecton_core.data_types import Float64Type
from tecton_core.data_types import Int32Type
from tecton_core.data_types import Int64Type
from tecton_core.data_types import StringType
from tecton_core.data_types import data_type_from_proto
from tecton_core.errors import TectonInternalError
from tecton_core.errors import TectonValidationError
from tecton_proto.common.calculation_node__client_pb2 import AbstractSyntaxTreeNode
from tecton_proto.common.calculation_node__client_pb2 import CaseStatement
from tecton_proto.common.calculation_node__client_pb2 import DatePart
from tecton_proto.common.calculation_node__client_pb2 import ElseClause
from tecton_proto.common.calculation_node__client_pb2 import LiteralValue
from tecton_proto.common.calculation_node__client_pb2 import Operation
from tecton_proto.common.calculation_node__client_pb2 import OperationType
from tecton_proto.common.calculation_node__client_pb2 import WhenClause
from tecton_proto.common.data_type__client_pb2 import DataType
from tecton_proto.common.data_type__client_pb2 import DataType as proto_DataType
from tecton_proto.common.data_type__client_pb2 import DataTypeEnum


# Matches numericDatatypesPrecisionOrder in go/internal/featureservice/featureengines/calculations/arithmetic.go
# Matches DuckDB precision hierarchy:
# See https://duckdb.org/docs/sql/data_types/numeric.html
# Float64 = DOUBLE
# Float32 = FLOAT
# Int32 = INTEGER
# Int64 = BIGINT
# Float64 + Float32 -> `SELECT 8::DOUBLE + 8::FLOAT;` returns DOUBLE, so Float64 is more precise than Float32
# Float32 + Int64 -> `select 8::FLOAT + 8::BIGINT;` returns FLOAT so Float32 is more precise than Int64
# Int64 + Int32 -> `select 8::BIGINT + 8::INTEGER;` returns BIGINT so Int64 is more precise than Int32
NUMERIC_DATATYPES_PRECISION_ORDER = [Float64Type().proto, Float32Type().proto, Int64Type().proto, Int32Type().proto]


class _ToAbstractSyntaxTree(Transformer):
    def __init__(self, original_expression: str, schema: Dict[str, SdkDataType]):
        self.original_expression = original_expression
        self.schema = schema
        super().__init__()

    def _type_is_numeric(self, input_dtype: proto_DataType) -> bool:
        return input_dtype in [Int64Type().proto, Int32Type().proto, Float64Type().proto, Float32Type().proto]

    def _type_is_boolean(self, input_dtype: proto_DataType):
        return input_dtype == BoolType().proto

    def _resolve_dtypes(self, input_dtypes: List[proto_DataType], operation: str) -> proto_DataType:
        if len(input_dtypes) == 0:
            return None
        all_numeric = all(self._type_is_numeric(input_dtype) for input_dtype in input_dtypes)
        if all_numeric:
            # check for presence of each numeric dtype in order, and return first one found
            for dtype in NUMERIC_DATATYPES_PRECISION_ORDER:
                if dtype in input_dtypes:
                    return dtype

        first_type = input_dtypes[0]
        for dtype in input_dtypes:
            if first_type != dtype:
                msg = f"Cannot combine datatypes {data_type_from_proto(first_type)} and {data_type_from_proto(dtype)} in operation {operation}"
                raise ExpressionParsingError(msg)
        # All dtypes are equal
        return first_type

    def _types_are_comparable(self, nodes: List[AbstractSyntaxTreeNode]) -> bool:
        if not nodes:
            # let's assume empty lists are comparable
            return True
        elif all(self._type_is_numeric(node.dtype) for node in nodes):
            # All numeric types are comparable (might change this later)
            return True
        elif any(node.dtype.type == DataTypeEnum.DATA_TYPE_STRUCT for node in nodes):
            # structs are not comparable
            return False
        else:
            # if all types are the same
            first_node_type = nodes[0].dtype
            return all(node.dtype == first_node_type for node in nodes)

    def literal(self, value) -> AbstractSyntaxTreeNode:
        value = value[0]
        if isinstance(value, bool):
            literal_value = LiteralValue(bool_value=value)
            dtype = BoolType()
        elif isinstance(value, int):
            literal_value = LiteralValue(int64_value=value)
            dtype = Int64Type()
        elif isinstance(value, float):
            literal_value = LiteralValue(float64_value=value)
            dtype = Float64Type()
        elif isinstance(value, str):
            literal_value = LiteralValue(string_value=value)
            dtype = StringType()
        else:
            msg = f"Something went wrong while parsing: Unexpected Literal type {type(value)}, node: {value}"
            raise RuntimeError(msg)
        return AbstractSyntaxTreeNode(literal_value=literal_value, dtype=dtype.proto)

    def numeric_negative(self, child: List[Token]):
        return -1 * child[0]

    def INT(self, value: str):
        return int(value)

    def DECIMAL(self, value: str):
        return float(value)

    def FLOAT(self, value: str):
        return float(value)

    def ESCAPED_STRING(self, value: str):
        value = value[1:-1]  # strip parentheses
        return value

    def SINGLE_QUOTED_STRING(self, value: str):
        value = value[1:-1]
        return value

    def true(self, value: str):
        return True

    def false(self, value: str):
        return False

    def day(self, value: str):
        return AbstractSyntaxTreeNode(date_part=DatePart.DAY)

    def month(self, value: str):
        return AbstractSyntaxTreeNode(date_part=DatePart.MONTH)

    def week(self, value: str):
        return AbstractSyntaxTreeNode(date_part=DatePart.WEEK)

    def year(self, value: str):
        return AbstractSyntaxTreeNode(date_part=DatePart.YEAR)

    def second(self, value: str):
        return AbstractSyntaxTreeNode(date_part=DatePart.SECOND)

    def hour(self, value: str):
        return AbstractSyntaxTreeNode(date_part=DatePart.HOUR)

    def minute(self, value: str):
        return AbstractSyntaxTreeNode(date_part=DatePart.MINUTE)

    def millennium(self, value: str):
        return AbstractSyntaxTreeNode(date_part=DatePart.MILLENNIUM)

    def century(self, value: str):
        return AbstractSyntaxTreeNode(date_part=DatePart.CENTURY)

    def decade(self, value: str):
        return AbstractSyntaxTreeNode(date_part=DatePart.DECADE)

    def quarter(self, value: str):
        return AbstractSyntaxTreeNode(date_part=DatePart.QUARTER)

    def milliseconds(self, value: str):
        return AbstractSyntaxTreeNode(date_part=DatePart.MILLISECONDS)

    def microseconds(self, value: str):
        return AbstractSyntaxTreeNode(date_part=DatePart.MICROSECONDS)

    def COLUMN_REFERENCE(self, value: str):
        # handle case where column reference is wrapped in "
        if value.startswith('"'):
            value = value[1:-1]

        if value not in self.schema:
            msg = f"Invalid column reference: {value}, possible values: {list(self.schema.keys())}"
            raise ExpressionParsingError(msg)
        dtype = self.schema[value]
        return AbstractSyntaxTreeNode(column_reference=value, dtype=dtype.tecton_type.proto)

    def coalesce(self, values) -> AbstractSyntaxTreeNode:
        if not values:
            # This shouldn't be possible because the regex pattern for this rule requires at least 1 argument,
            # but adding as extra protection in case the rule changes in the future.
            msg = f"Error parsing expression: At least one value required for COALESCE, {self.original_expression}"
            raise ExpressionParsingError(msg)

        dtypes = {data_type_from_proto(value.dtype) for value in values}
        if len(dtypes) != 1:
            msg = f"Cannot mix values of types {dtypes} in COALESCE operator."
            raise ExpressionParsingError(msg)

        return AbstractSyntaxTreeNode(
            operation=Operation(operation=OperationType.COALESCE, operands=values), dtype=values[0].dtype
        )

    def datediff(self, values) -> AbstractSyntaxTreeNode:
        if not values or len(values) != 3:
            msg = f"Error parsing expression: Exactly three values required for DATEDIFF, {self.original_expression}"
            raise ExpressionParsingError(msg)

        date_part = values[0]
        start_date = values[1]
        end_date = values[2]
        if (
            start_date.dtype.type != DataTypeEnum.DATA_TYPE_TIMESTAMP
            or end_date.dtype.type != DataTypeEnum.DATA_TYPE_TIMESTAMP
        ):
            msg = "For DATEDIFF, both start_date and end_date must have type TIMESTAMP."
            raise ExpressionParsingError(msg)

        operands = [date_part, start_date, end_date]
        return AbstractSyntaxTreeNode(
            operation=Operation(operation=OperationType.DATE_DIFF, operands=operands), dtype=Int64Type().proto
        )

    def _comparison(self, values: List[AbstractSyntaxTreeNode], operation) -> AbstractSyntaxTreeNode:
        if len(values) != 2:
            # Shouldn't be possible due to regex patterns, so this is an internal error
            msg = f"Comparisons should have exactly 2 operands, found {len(values)}"
            raise TectonInternalError(msg)
        if not self._types_are_comparable(values):
            left_type = data_type_from_proto(values[0].dtype)
            right_type = data_type_from_proto(values[1].dtype)
            msg = f"Types {left_type} and {right_type} are not comparable"
            raise ExpressionParsingError(msg)
        return AbstractSyntaxTreeNode(operation=Operation(operation=operation, operands=values), dtype=BoolType().proto)

    def equals(self, values) -> AbstractSyntaxTreeNode:
        return self._comparison(values, OperationType.EQUALS)

    def not_equals(self, values) -> AbstractSyntaxTreeNode:
        return self._comparison(values, OperationType.NOT_EQUALS)

    def greater_than(self, values) -> AbstractSyntaxTreeNode:
        return self._comparison(values, OperationType.GREATER_THAN)

    def greater_than_or_equals(self, values) -> AbstractSyntaxTreeNode:
        return self._comparison(values, OperationType.GREATER_THAN_EQUALS)

    def less_than(self, values) -> AbstractSyntaxTreeNode:
        return self._comparison(values, OperationType.LESS_THAN)

    def less_than_or_equals(self, values) -> AbstractSyntaxTreeNode:
        return self._comparison(values, OperationType.LESS_THAN_EQUALS)

    def _determine_math_dtype(self, values: List[AbstractSyntaxTreeNode], operation: OperationType) -> DataType:
        for value in values:
            if not self._type_is_numeric(value.dtype):
                msg = f"Cannot perform {OperationType.Name(operation)} on non-numeric type {data_type_from_proto(value.dtype)}"
                raise ExpressionParsingError(msg)

        if operation == OperationType.DIVISION:
            return Float64Type().proto

        value_dtypes = [value.dtype for value in values]
        return self._resolve_dtypes(value_dtypes, OperationType.Name(operation))

    def _math_operation(self, values, operation):
        dtype_proto = self._determine_math_dtype(values, operation)
        return AbstractSyntaxTreeNode(operation=Operation(operation=operation, operands=values), dtype=dtype_proto)

    def addition(self, values) -> AbstractSyntaxTreeNode:
        return self._math_operation(values, OperationType.ADDITION)

    def subtraction(self, values) -> AbstractSyntaxTreeNode:
        return self._math_operation(values, OperationType.SUBTRACTION)

    def multiplication(self, values) -> AbstractSyntaxTreeNode:
        return self._math_operation(values, OperationType.MULTIPLICATION)

    def division(self, values) -> AbstractSyntaxTreeNode:
        return self._math_operation(values, OperationType.DIVISION)

    def negation(self, value):
        neg_1 = AbstractSyntaxTreeNode(literal_value=LiteralValue(int64_value=-1), dtype=Int32Type().proto)
        return self.multiplication([neg_1, value[0]])

    def searched_case_statement(self, clauses: Union[WhenClause, ElseClause]):
        if not clauses or not isinstance(clauses[0], WhenClause):
            msg = "Case statements require at least 1 WHEN clause."
            raise ExpressionParsingError(msg)
        statement_dtype_proto = self._resolve_dtypes([clause.result.dtype for clause in clauses], "CASE statement")

        if isinstance(clauses[-1], ElseClause):
            else_clause = clauses.pop()
        else:
            else_clause = None
        when_clauses = clauses
        for when_clause in when_clauses:
            condition_dtype = data_type_from_proto(when_clause.condition.dtype)
            # Only these types are acceptable types in a when-statement conditional
            if not isinstance(condition_dtype, (Int64Type, Int32Type, BoolType, Float64Type, Float32Type)):
                msg = f"Could not convert CASE STATEMENT predicate type {condition_dtype} to BOOL."
                raise ExpressionParsingError(msg)

        return AbstractSyntaxTreeNode(
            case_statement=CaseStatement(when_clauses=when_clauses, else_clause=else_clause),
            dtype=statement_dtype_proto,
        )

    def when_clause(self, values):
        return WhenClause(condition=values[0], result=values[1])

    def else_clause(self, value):
        return ElseClause(result=value[0])

    def logical_not(self, values) -> AbstractSyntaxTreeNode:
        return self._logical_operator(values, OperationType.NOT)

    def logical_and(self, values) -> AbstractSyntaxTreeNode:
        return self._logical_operator(values, OperationType.AND)

    def logical_or(self, values) -> AbstractSyntaxTreeNode:
        return self._logical_operator(values, OperationType.OR)

    def _logical_operator(
        self, values: List[AbstractSyntaxTreeNode], operation_type: OperationType
    ) -> AbstractSyntaxTreeNode:
        for value in values:
            if not (self._type_is_boolean(value.dtype) or self._type_is_numeric(value.dtype)):
                msg = f"{OperationType.Name(operation_type)} operator was evaluated with value of type {data_type_from_proto(value.dtype)} which could not be converted to bool."
                raise ExpressionParsingError(msg)
        return AbstractSyntaxTreeNode(
            operation=Operation(operation=operation_type, operands=values), dtype=BoolType().proto
        )


class CustomParseError(ParseError):
    def __init__(self, matching_regexes: []):
        super().__init__()
        self._regexes = matching_regexes

    @property
    def matching_regexes(self):
        return self._regexes

    @property
    def error_message(self):
        raise NotImplementedError


class DateDiffParseError(CustomParseError):
    def __init__(self):
        super().__init__(matching_regexes=[r"datediff\s*\(\s*\""])

    @property
    def error_message(self):
        return "Error parsing expression: the date part argument to DATEDIFF requires single quotes (') not double quotes (\"). For example, instead of datediff(\"day\",...), use datediff('day',...)."


CUSTOM_PARSE_ERRORS = [DateDiffParseError()]


def expression_to_proto(expression: str, schema: Dict[str, SdkDataType]) -> AbstractSyntaxTreeNode:
    # TODO: This class initialization should be top-level for better performance, but for not that causes
    #  issues with type-hint checking. We should fix that.
    lexer = Lark_StandAlone()
    try:
        lark_syntax_tree = lexer.parse(expression)
    except (LexError, ParseError) as e:
        for custom_parse_error in CUSTOM_PARSE_ERRORS:
            if any(re.search(matching_regex, expression) for matching_regex in custom_parse_error.matching_regexes):
                msg = custom_parse_error.error_message
                raise ExpressionParsingError(msg) from e
        msg = f"Error parsing expression: {expression}"
        raise ExpressionParsingError(msg) from e
    try:
        ast = _ToAbstractSyntaxTree(expression, schema).transform(lark_syntax_tree)
    except VisitError as e:
        if isinstance(e.orig_exc, TectonValidationError):
            raise e.orig_exc  # Re-raise the original TectonValidationError
        else:
            # This indicates an issue with our parsing logic, not a user error.
            msg = f"Error converting parse tree to AST: {lark_syntax_tree}"
            raise TectonInternalError(msg) from e
    return ast
